﻿/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Learners
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Reflection;
    using System.Runtime.Serialization;
    using System.Runtime.Serialization.Formatters.Binary;
    using System.Text;

    /// <summary>
    /// Implements various utilities for all learners.
    /// </summary>
    public static class Utilities
    {
        #region Serialization

        #region Save

        /// <summary>
        /// Persists a learner to a file.
        /// </summary>
        /// <param name="learner">The learner to serialize.</param>
        /// <param name="fileName">The name of the file.</param>
        public static void Save(this ILearner learner, string fileName)
        {
            if (learner == null)
            {
                throw new ArgumentNullException("learner");
            }

            if (fileName == null)
            {
                throw new ArgumentNullException("fileName");
            }

            using (Stream stream = File.Open(fileName, FileMode.Create))
            {
                var formatter = new BinaryFormatter();
                learner.Save(stream, formatter);
            }
        }

        /// <summary>
        /// Serializes a learner to a given stream using a given formatter.
        /// </summary>
        /// <param name="learner">The learner.</param>
        /// <param name="stream">The serialization stream.</param>
        /// <param name="formatter">The formatter.</param>
        public static void Save(this ILearner learner, Stream stream, IFormatter formatter)
        {
            if (learner == null)
            {
                throw new ArgumentNullException("learner");
            }

            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }

            if (formatter == null)
            {
                throw new ArgumentNullException("formatter");
            }

            AppendVersionMetadata(learner.GetType(), stream, formatter);
            formatter.Serialize(stream, learner);
        }

        /// <summary>
        /// Persists an object that controls its binary serialization to a file with the specified name.
        /// </summary>
        /// <param name="obj">The object to be serialized.</param>
        /// <param name="fileName">The name of the file.</param>
        /// <param name="fileMode">The mode used to open the file. Defaults to <see cref="FileMode.Create"/>.</param>
        /// <remarks>To load a saved learner, you can use the factory methods whose names start with LoadBackwardCompatible.</remarks>
        public static void SaveForwardCompatible(this ICustomSerializable obj, string fileName, FileMode fileMode = FileMode.Create)
        {
            if (obj == null)
            {
                throw new ArgumentNullException("obj");
            }

            if (fileName == null)
            {
                throw new ArgumentNullException("fileName");
            }

            using (Stream stream = File.Open(fileName, fileMode))
            {
                obj.SaveForwardCompatible(stream);
            }
        }

        /// <summary>
        /// Persists an object that controls its binary serialization to the specified stream.
        /// </summary>
        /// <param name="obj">The object to be serialized.</param>
        /// <param name="stream">The serialization stream.</param>
        /// <remarks>To load a saved learner, you can use the factory methods whose names start with LoadBackwardCompatible.</remarks>
        public static void SaveForwardCompatible(this ICustomSerializable obj, Stream stream)
        {
            if (obj == null)
            {
                throw new ArgumentNullException("obj");
            }

            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }

            using (var writer = new BinaryWriter(stream, Encoding.UTF8, true))
            {
                obj.SaveForwardCompatible(writer);
            }
        }

        #endregion

        #region Load

        /// <summary>
        /// Deserializes a learner from a given stream and formatter.
        /// </summary>
        /// <typeparam name="TLearner">The type of a learner.</typeparam>
        /// <param name="stream">The stream.</param>
        /// <param name="formatter">The formatter.</param>
        /// <returns>The deserialized learner object.</returns>
        public static TLearner Load<TLearner>(Stream stream, IFormatter formatter)
        {
            if (stream == null)
            {
                throw new ArgumentNullException("stream");
            }

            if (formatter == null)
            {
                throw new ArgumentNullException("formatter");
            }

            CheckVersion(stream, formatter);
            return (TLearner)formatter.Deserialize(stream);
        }

        /// <summary>
        /// Deserializes a learner from a file.
        /// </summary>
        /// <typeparam name="TLearner">The type of a learner.</typeparam>
        /// <param name="fileName">The file name.</param>
        /// <returns>The deserialized learner object.</returns>
        public static TLearner Load<TLearner>(string fileName)
        {
            if (fileName == null)
            {
                throw new ArgumentNullException("fileName");
            }

            using (Stream stream = File.Open(fileName, FileMode.Open))
            {
                var formatter = new BinaryFormatter();
                return Load<TLearner>(stream, formatter);
            }
        }

        #endregion

        #endregion

        #region Batching

        /// <summary>
        /// Gets the specified batch of elements from a given array.
        /// </summary>
        /// <typeparam name="T">The type of an array element.</typeparam>
        /// <param name="batch">The zero-indexed batch to get the array elements for.</param>
        /// <param name="array">An array to get the batch of elements from.</param>
        /// <param name="totalBatchCount">The total number of batches.</param>
        /// <returns>The array elements for the specified batch.</returns>
        public static IList<T> GetBatch<T>(int batch, T[] array, int totalBatchCount)
        {
            if (array == null)
            {
                throw new ArgumentNullException("array");
            }

            if (totalBatchCount < 1)
            {
                throw new ArgumentOutOfRangeException("totalBatchCount", "The total number of batches must be greater than 0.");
            }

            if (batch < 0)
            {
                throw new ArgumentOutOfRangeException("batch", "The requested batch number must not be negative.");
            }

            if (batch >= totalBatchCount)
            {
                throw new ArgumentOutOfRangeException("batch", "The requested batch number must not be greater than the total number of batches.");
            }

            // Avoid the batching performance overhead when the data is not batched.
            if (array.Length == 0)
            {
                return array;
            }

            if (totalBatchCount == 1)
            {
                return array;
            }

            if (array.Length < totalBatchCount)
            {
                throw new ArgumentException("There are not enough elements for the requested number of batches.");
            }

            // Compute start and size of requested batch
            int batchStart = 0;
            int batchSize = 0;
            int instanceCount = array.Length;
            for (int b = 0; b <= batch; b++)
            {
                batchStart += batchSize;
                batchSize = (int)Math.Ceiling((double)instanceCount / (totalBatchCount - b));
                instanceCount -= batchSize;
            }

            return new ArraySegment<T>(array, batchStart, batchSize);
        }

        #endregion

        #region Helper methods

        /// <summary>
        /// Appends learner metadata required for version checking to a given serialization stream.
        /// </summary>
        /// <param name="type">The type of the learner.</param>
        /// <param name="stream">The serialization stream.</param>
        /// <param name="formatter">The formatter.</param>
        private static void AppendVersionMetadata(Type type, Stream stream, IFormatter formatter)
        {
            formatter.Serialize(stream, type);
            formatter.Serialize(stream, GetSerializationVersion(type));
        }

        /// <summary>
        /// Checks the current version of a learner matches the one in a given serialization stream.
        /// </summary>
        /// <param name="stream">The stream containing the required metadata.</param>
        /// <param name="formatter">The formatter.</param>
        /// <remarks>This method modifies the given stream.</remarks>
        private static void CheckVersion(Stream stream, IFormatter formatter)
        {
            if (formatter == null)
            {
                throw new ArgumentNullException("formatter");
            }

            var type = (Type)formatter.Deserialize(stream);
            var expectedSerializationVersion = GetSerializationVersion(type);

            var actualSerializationVersion = (int)formatter.Deserialize(stream);

            if (expectedSerializationVersion != actualSerializationVersion)
            {
                throw new SerializationException(
                    string.Format(
                        "Serialization version mismatch. Expected: {0}, actual: {1}.",
                        expectedSerializationVersion,
                        actualSerializationVersion));
            }
        }

        /// <summary>
        /// Gets the serialization version of a learner.
        /// </summary>
        /// <param name="memberInfo">The member info of the learner.</param>
        /// <returns>The serialization version of the learner.</returns>
        private static int GetSerializationVersion(MemberInfo memberInfo)
        {
            var serializationVersionAttribute = 
                (SerializationVersionAttribute)Attribute.GetCustomAttribute(memberInfo, typeof(SerializationVersionAttribute));

            if (serializationVersionAttribute == null)
            {
                throw new SerializationException(
                    string.Format(
                        "The {0} must be applied to the learner for serialization and deserialization.",
                        typeof(SerializationVersionAttribute).Name));
            }

            return serializationVersionAttribute.SerializationVersion;
        }

        #endregion
    }
}
