﻿/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Learners.MatchboxRecommenderInternal
{
    using System;
    using System.Diagnostics;
    using System.IO;

    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Utils;

    using GaussianArray = Distributions.DistributionStructArray<MicrosoftResearch.Infer.Distributions.Gaussian, double>;

    /// <summary>
    /// Represents distribution over traits and bias of a user or an item.
    /// </summary>
    [Serializable]
    internal abstract class EntityParameterDistribution : ICustomSerializable
    {
        /// <summary>
        /// The current custom binary serialization version of the <see cref="EntityParameterDistribution"/> class.
        /// </summary>
        private const int CustomSerializationVersion = 1;

        /// <summary>
        /// Initializes a new instance of the <see cref="EntityParameterDistribution"/> class
        /// from a reader of a binary stream.
        /// </summary>
        /// <param name="reader">The binary reader to read the distribution over traits and bias of a user or an item from.</param>
        protected EntityParameterDistribution(BinaryReader reader)
        {
            Debug.Assert(reader != null, "The reader must not be null.");

            int deserializedVersion = reader.ReadSerializationVersion(CustomSerializationVersion);

            if (deserializedVersion == CustomSerializationVersion)
            {
                this.Traits = reader.ReadGaussianArray();
                this.Bias = reader.ReadGaussian();
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="EntityParameterDistribution"/> class.
        /// </summary>
        /// <param name="traitDistribution">The distribution over the entity traits.</param>
        /// <param name="biasDistribution">The distribution over the entity bias.</param>
        protected EntityParameterDistribution(GaussianArray traitDistribution, Gaussian biasDistribution)
        {
            Debug.Assert(traitDistribution != null, "A valid distribution over traits must be provided.");
            
            this.Traits = traitDistribution;
            this.Bias = biasDistribution;
        }

        /// <summary>
        /// Gets the distribution over the traits.
        /// </summary>
        public GaussianArray Traits { get; private set; }

        /// <summary>
        /// Gets the distribution over the bias.
        /// </summary>
        public Gaussian Bias { get; private set; }

        /// <summary>
        /// Saves the distribution over traits and bias of a user or an item using the specified writer to a binary stream.
        /// </summary>
        /// <param name="writer">The writer to save the distribution over traits and bias of a user or an item to.</param>
        public virtual void SaveForwardCompatible(BinaryWriter writer)
        {
            writer.Write(CustomSerializationVersion);

            writer.Write(this.Traits);
            writer.Write(this.Bias);
        }
    }
}
