/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Learners.MatchboxRecommenderInternal
{
    using System;
    using System.Diagnostics;
    using System.IO;
    using System.Linq;

    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Utils;

    using GaussianArray = Distributions.DistributionStructArray<MicrosoftResearch.Infer.Distributions.Gaussian, double>;
    using GaussianMatrix = Distributions.DistributionRefArray<Distributions.DistributionStructArray<MicrosoftResearch.Infer.Distributions.Gaussian, double>, double[]>;

    /// <summary>
    /// Represents the distribution over feature weights.
    /// </summary>
    [Serializable]
    internal class FeatureParameterDistribution : ICloneable, ICustomSerializable
    {
        /// <summary>
        /// The current custom binary serialization version of the <see cref="FeatureParameterDistribution"/> class.
        /// </summary>
        private const int CustomSerializationVersion = 1;

        /// <summary>
        /// Initializes a new instance of the <see cref="FeatureParameterDistribution"/> class.
        /// </summary>
        public FeatureParameterDistribution()
        {
            this.TraitWeights = new GaussianMatrix(0);
            this.BiasWeights = new GaussianArray(0);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FeatureParameterDistribution"/> class
        /// from a reader of a binary stream.
        /// </summary>
        /// <param name="reader">The binary reader to read the distribution over feature weights from.</param>
        public FeatureParameterDistribution(BinaryReader reader)
        {
            Debug.Assert(reader != null, "The reader must not be null.");

            int deserializedVersion = reader.ReadSerializationVersion(CustomSerializationVersion);

            if (deserializedVersion == CustomSerializationVersion)
            {
                this.TraitWeights = reader.ReadGaussianMatrix();
                this.BiasWeights = reader.ReadGaussianArray();
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FeatureParameterDistribution"/> class.
        /// </summary>
        /// <param name="traitFeatureWeightDistribution">The distribution over weights of the feature contribution to traits.</param>
        /// <param name="biasFeatureWeightDistribution">The distribution over weights of the feature contribution to biases.</param>
        public FeatureParameterDistribution(GaussianMatrix traitFeatureWeightDistribution, GaussianArray biasFeatureWeightDistribution)
        {
            Debug.Assert(
                (traitFeatureWeightDistribution == null && biasFeatureWeightDistribution == null) ||
                traitFeatureWeightDistribution.All(w => w != null && w.Count == biasFeatureWeightDistribution.Count),
                "The provided distributions should be valid and consistent in the number of features.");
            
            this.TraitWeights = traitFeatureWeightDistribution;
            this.BiasWeights = biasFeatureWeightDistribution;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FeatureParameterDistribution"/> class.
        /// </summary>
        /// <param name="traitCount">The number of traits.</param>
        /// <param name="featureCount">The number of features.</param>
        /// <param name="value">
        /// The value to which each element of the 
        /// contained distributions will be initialized.
        /// </param>
        public FeatureParameterDistribution(int traitCount, int featureCount, Gaussian value)
        {
            this.TraitWeights = new GaussianMatrix(new GaussianArray(value, featureCount), traitCount);
            this.BiasWeights = new GaussianArray(value, featureCount);
        }

        /// <summary>
        /// Gets the number of features.
        /// </summary>
        public int FeatureCount
        {
            get { return this.BiasWeights.Count; }
        }

        /// <summary>
        /// Gets the distribution over weights of the feature contribution to traits.
        /// </summary>
        public GaussianMatrix TraitWeights { get; private set; }

        /// <summary>
        /// Gets the distribution over weights of the feature contribution to biases.
        /// </summary>
        public GaussianArray BiasWeights { get; private set; }

        /// <summary>
        /// Creates a new object that is a copy of the current instance.
        /// </summary>
        /// <returns>A new object that is a copy of this instance.</returns>
        public object Clone()
        {
            return new FeatureParameterDistribution
            {
                TraitWeights = (GaussianMatrix)this.TraitWeights.Clone(),
                BiasWeights = (GaussianArray)this.BiasWeights.Clone(),
            };
        }

        /// <summary>
        /// Saves the distributions over feature weights using the specified writer to a binary stream.
        /// </summary>
        /// <param name="writer">The writer to save the distributions over feature weights to.</param>
        public void SaveForwardCompatible(BinaryWriter writer)
        {
            writer.Write(CustomSerializationVersion);
            writer.Write(this.TraitWeights);
            writer.Write(this.BiasWeights);
        }
    }
}