/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

    /// <summary>Provides outgoing messages for <see cref="Factor.ArrayFromVector(Vector)" />, given random arguments to the function.</summary>
    [FactorMethod(typeof(Factor), "ArrayFromVector")]
    [Quality(QualityBand.Stable)]
    public static class ArrayFromVectorOp
    {
        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Constant value for <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(array,vector))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(double[] array, Vector vector)
        {
            if (array.Length != vector.Count)
                return Double.NegativeInfinity;
            for (int i = 0; i < array.Length; i++)
            {
                if (array[i] != vector[i])
                    return Double.NegativeInfinity;
            }
            return 0.0;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Constant value for <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(array,vector))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(double[] array, Vector vector)
        {
            return LogAverageFactor(array, vector);
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Constant value for <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(array,vector))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        public static double AverageLogFactor(double[] array, Vector vector)
        {
            return LogAverageFactor(array, vector);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(vector) p(vector) factor(array,vector))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(double[] array, VectorGaussian vector)
        {
            return vector.GetLogProb(Vector.FromArray(array));
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(vector) p(vector) factor(array,vector))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(double[] array, VectorGaussian vector)
        {
            return LogAverageFactor(array, vector);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(vector) p(vector) factor(array,vector))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(double[] array, VectorGaussianMoments vector)
        {
            return vector.GetLogProb(Vector.FromArray(array));
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(vector) p(vector) factor(array,vector))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(double[] array, VectorGaussianMoments vector)
        {
            return LogAverageFactor(array, vector);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <param name="to_vector">Outgoing message to <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(array,vector) p(array,vector) factor(array,vector))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(IList<Gaussian> array, VectorGaussian vector, [Fresh] VectorGaussian to_vector)
        {
            return vector.GetLogAverageOf(to_vector);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <param name="to_vector">Outgoing message to <c>vector</c>.</param>
        /// <param name="to_array">Outgoing message to <c>array</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(array,vector) p(array,vector) factor(array,vector) / sum_array p(array) messageTo(array))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(IList<Gaussian> array, VectorGaussian vector, [Fresh] VectorGaussian to_vector, [Fresh] IList<Gaussian> to_array)
        {
            double result = LogAverageFactor(array, vector, to_vector);
            int length = array.Count;
            for (int i = 0; i < length; i++)
            {
                result -= to_array[i].GetLogAverageOf(array[i]);
            }
            return result;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <param name="to_vector">Outgoing message to <c>vector</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(array,vector) p(array,vector) factor(array,vector))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(IList<Gaussian> array, VectorGaussianMoments vector, [Fresh] VectorGaussianMoments to_vector)
        {
            return vector.GetLogAverageOf(to_vector);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <param name="to_vector">Outgoing message to <c>vector</c>.</param>
        /// <param name="to_array">Outgoing message to <c>array</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(array,vector) p(array,vector) factor(array,vector) / sum_array p(array) messageTo(array))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(IList<Gaussian> array, VectorGaussianMoments vector, [Fresh] VectorGaussianMoments to_vector, [Fresh] IList<Gaussian> to_array)
        {
            double result = LogAverageFactor(array, vector, to_vector);
            int length = array.Count;
            for (int i = 0; i < length; i++)
            {
                result -= to_array[i].GetLogAverageOf(array[i]);
            }
            return result;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(double[] array, VectorGaussian vector)
        {
            return 0.0;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(IList<Gaussian> array, VectorGaussian vector)
        {
            return 0.0;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(double[] array, VectorGaussianMoments vector)
        {
            return 0.0;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(IList<Gaussian> array, VectorGaussianMoments vector)
        {
            return 0.0;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="array">Incoming message from <c>array</c>.</param>
        /// <param name="vector">Constant value for <c>vector</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(IList<Gaussian> array, Vector vector)
        {
            return 0.0;
        }

        /// <summary>EP message to <c>vector</c>.</summary>
        /// <param name="array">Incoming message from <c>array</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>vector</c> as the random arguments are varied. The formula is <c>proj[p(vector) sum_(array) p(array) factor(array,vector)]/p(vector)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="array" /> is not a proper distribution.</exception>
        public static VectorGaussian VectorAverageConditional([SkipIfAllUniform] IList<Gaussian> array, VectorGaussian result)
        {
            return VectorAverageLogarithm(array, result);
        }

        /// <summary>VMP message to <c>vector</c>.</summary>
        /// <param name="array">Incoming message from <c>array</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> with <c>array</c> integrated out. The formula is <c>sum_array p(array) factor(array,vector)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="array" /> is not a proper distribution.</exception>
        public static VectorGaussian VectorAverageLogarithm([SkipIfAllUniform] IList<Gaussian> array, VectorGaussian result)
        {
            if (result.Dimension != array.Count)
                throw new ArgumentException("array.Count (" + array.Count + ") != result.Dimension (" + result.Dimension + ")");
            result.Precision.SetAllElementsTo(0.0);
            int length = array.Count;
            for (int i = 0; i < length; i++)
            {
                Gaussian item = array[i];
                result.Precision[i, i] = item.Precision;
                result.MeanTimesPrecision[i] = item.MeanTimesPrecision;
            }
            return result;
        }

        /// <summary>EP message to <c>vector</c>.</summary>
        /// <param name="array">Incoming message from <c>array</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>vector</c> as the random arguments are varied. The formula is <c>proj[p(vector) sum_(array) p(array) factor(array,vector)]/p(vector)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="array" /> is not a proper distribution.</exception>
        public static VectorGaussianMoments VectorAverageConditional([SkipIfAllUniform] IList<Gaussian> array, VectorGaussianMoments result)
        {
            return VectorAverageLogarithm(array, result);
        }

        /// <summary>VMP message to <c>vector</c>.</summary>
        /// <param name="array">Incoming message from <c>array</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> with <c>array</c> integrated out. The formula is <c>sum_array p(array) factor(array,vector)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="array" /> is not a proper distribution.</exception>
        public static VectorGaussianMoments VectorAverageLogarithm([SkipIfAllUniform] IList<Gaussian> array, VectorGaussianMoments result)
        {
            if (result.Dimension != array.Count)
                throw new ArgumentException("array.Count (" + array.Count + ") != result.Dimension (" + result.Dimension + ")");
            result.Variance.SetAllElementsTo(0.0);
            int length = array.Count;
            for (int i = 0; i < length; i++)
            {
                Gaussian item = array[i];
                double m, v;
                item.GetMeanAndVariance(out m, out v);
                result.Variance[i, i] = v;
                result.Mean[i] = m;
            }
            return result;
        }

        /// <summary>EP message to <c>vector</c>.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussian VectorAverageConditional(double[] array, VectorGaussian result)
        {
            return VectorAverageLogarithm(array, result);
        }

        /// <summary>VMP message to <c>vector</c>.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussian VectorAverageLogarithm(double[] array, VectorGaussian result)
        {
            result.Point = result.Point;
            result.Point.SetTo(array);
            return result;
        }

        /// <summary>EP message to <c>vector</c>.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussianMoments VectorAverageConditional(double[] array, VectorGaussianMoments result)
        {
            return VectorAverageLogarithm(array, result);
        }

        /// <summary>VMP message to <c>vector</c>.</summary>
        /// <param name="array">Constant value for <c>array</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>vector</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussianMoments VectorAverageLogarithm(double[] array, VectorGaussianMoments result)
        {
            result.Point = result.Point;
            result.Point.SetTo(array);
            return result;
        }

        /// <typeparam name="GaussianList">The type of the resulting array.</typeparam>
        public static GaussianList ArrayAverageConditional<GaussianList>(
            [NoInit] IList<Gaussian> array, [SkipIfUniform] VectorGaussian vector, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            if (result.Count != vector.Dimension)
                throw new ArgumentException("vector.Dimension (" + vector.Dimension + ") != result.Count (" + result.Count + ")");
            int length = result.Count;
            bool allPointMass = array.All(g => g.IsPointMass);
            if (allPointMass)
            {
                // efficient special case
                for (int i = 0; i < length; i++)
                {
                    double x = array[i].Point;
                    // -prec*(x-m) = -prec*x + prec*m
                    double dlogp = vector.MeanTimesPrecision[i];
                    for (int j = 0; j < length; j++)
                    {
                        dlogp -= vector.Precision[i, j] * array[j].Point;
                    }
                    double ddlogp = -vector.Precision[i, i];
                    result[i] = Gaussian.FromDerivatives(x, dlogp, ddlogp, false);
                }
            }
            else if (vector.IsPointMass)
            {
                // efficient special case
                Vector mean = vector.Point;
                for (int i = 0; i < length; i++)
                {
                    result[i] = Gaussian.PointMass(mean[i]);
                }
            }
            else if (vector.IsUniform())
            {
                for (int i = 0; i < length; i++)
                {
                    result[i] = Gaussian.Uniform();
                }
            }
            else if (array.Any(g => g.IsPointMass))
            {
                // Z = N(m1; m2, V1+V2)
                // logZ = -0.5 (m1-m2)'inv(V1+V2)(m1-m2)
                // dlogZ = (m1-m2)'inv(V1+V2) dm2
                // ddlogZ = -dm2'inv(V1+V2) dm2
                Vector mean = Vector.Zero(length);
                PositiveDefiniteMatrix variance = new PositiveDefiniteMatrix(length, length);
                vector.GetMeanAndVariance(mean, variance);
                for (int i = 0; i < length; i++)
                {
                    if (array[i].IsUniform()) continue;
                    double m, v;
                    array[i].GetMeanAndVariance(out m, out v);
                    variance[i, i] += v;
                    mean[i] -= m;
                }
                PositiveDefiniteMatrix precision = variance.Inverse();
                Vector meanTimesPrecision = precision * mean;
                for (int i = 0; i < length; i++)
                {
                    if (array[i].IsUniform())
                    {
                        result[i] = Gaussian.FromMeanAndVariance(mean[i], variance[i, i]);
                    }
                    else
                    {
                        double alpha = meanTimesPrecision[i];
                        double beta = precision[i, i];
                        result[i] = GaussianOp.GaussianFromAlphaBeta(array[i], alpha, beta, false);
                    }
                }
            }
            else
            {
                // Compute inv(V1+V2)*(m1-m2) as inv(V2)*inv(inv(V1) + inv(V2))*(inv(V1)*m1 + inv(V2)*m2) - inv(V2)*m2 = inv(V2)*(m - m2)
                // Compute inv(V1+V2) as inv(V2)*inv(inv(V1) + inv(V2))*inv(V2) - inv(V2)
                PositiveDefiniteMatrix precision = (PositiveDefiniteMatrix)vector.Precision.Clone();
                Vector meanTimesPrecision = vector.MeanTimesPrecision.Clone();
                for (int i = 0; i < length; i++)
                {
                    Gaussian g = array[i];
                    precision[i, i] += g.Precision;
                    meanTimesPrecision[i] += g.MeanTimesPrecision;
                }
                bool fastMethod = true;
                if (fastMethod)
                {
                    bool isPosDef;
                    // this destroys precision
                    LowerTriangularMatrix precisionChol = precision.CholeskyInPlace(out isPosDef);
                    if (!isPosDef) throw new PositiveDefiniteMatrixException();
                    // variance = inv(precisionChol*precisionChol') = inv(precisionChol)'*inv(precisionChol) = varianceChol*varianceChol'
                    // this destroys meanTimesPrecision
                    var mean = meanTimesPrecision.PredivideBy(precisionChol);
                    mean = mean.PredivideByTranspose(precisionChol);
                    var varianceCholTranspose = precisionChol;
                    // this destroys precisionChol
                    varianceCholTranspose.SetToInverse(precisionChol);
                    for (int i = 0; i < length; i++)
                    {
                        Gaussian g = array[i];
                        double variance_ii = GetSquaredLengthOfColumn(varianceCholTranspose, i);
                        // works when g is uniform, but not when g is point mass
                        result[i] = Gaussian.FromMeanAndVariance(mean[i], variance_ii) / g;
                    }
                }
                else
                {
                    // equivalent to above, but slower
                    PositiveDefiniteMatrix variance = precision.Inverse();
                    var mean = variance * meanTimesPrecision;
                    for (int i = 0; i < length; i++)
                    {
                        Gaussian g = array[i];
                        // works when g is uniform, but not when g is point mass
                        result[i] = Gaussian.FromMeanAndVariance(mean[i], variance[i, i]) / g;
                    }
                }
            }
            return result;
        }

        private static double GetSquaredLengthOfRow(LowerTriangularMatrix L, int row)
        {
            double sum = 0;
            for (int col = 0; col <= row; col++)
            {
                double element = L[row, col];
                sum += element * element;
            }
            return sum;
        }

        private static double GetSquaredLengthOfColumn(LowerTriangularMatrix L, int column)
        {
            int rows = L.Rows;
            double sum = 0;
            for (int row = column; row < rows; row++)
            {
                double element = L[row, column];
                sum += element * element;
            }
            return sum;
        }

        /// <summary />
        /// <param name="vector">Incoming message from <c>vector</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static DistributionStructArray<Gaussian, double> ArrayAverageConditionalInit([IgnoreDependency] VectorGaussian vector)
        {
            return new DistributionStructArray<Gaussian, double>(vector.Dimension);
        }

        /// <typeparam name="GaussianList">The type of the resulting array.</typeparam>
        public static GaussianList ArrayAverageConditional<GaussianList>(
            [NoInit] IList<Gaussian> array, [SkipIfUniform] VectorGaussianMoments vector, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            if (result.Count != vector.Dimension)
                throw new ArgumentException("vector.Dimension (" + vector.Dimension + ") != result.Count (" + result.Count + ")");
            int length = result.Count;
            if (array.All(g => g.IsUniform()) || vector.IsUniform() || vector.IsPointMass)
            {
                for (int i = 0; i < length; i++)
                {
                    result[i] = Gaussian.FromMeanAndVariance(vector.Mean[i], vector.Variance[i, i]);
                }
            }
            else
            {
                // Z = N(m1; m2, V1+V2)
                // logZ = -0.5 (m1-m2)'inv(V1+V2)(m1-m2)
                // dlogZ = (m1-m2)'inv(V1+V2) dm2
                // ddlogZ = -dm2'inv(V1+V2) dm2
                VectorGaussianMoments product = new VectorGaussianMoments(length);
                Vector mean = product.Mean;
                PositiveDefiniteMatrix variance = product.Variance;
                vector.GetMeanAndVariance(mean, variance);
                for (int i = 0; i < length; i++)
                {
                    double m, v;
                    array[i].GetMeanAndVariance(out m, out v);
                    variance[i, i] += v;
                    mean[i] -= m;
                }
                double offset = 1e-10;
                for (int i = 0; i < length; i++)
                {
                    variance[i, i] += offset;
                }
                PositiveDefiniteMatrix precision = variance.Inverse();
                Vector meanTimesPrecision = precision * mean;
                double precisionThreshold = 1;
                if (array.Any(g => g.Precision <= precisionThreshold))
                {
                    // compute the posterior mean and variance
                    product.SetToProduct(vector, array);
                    mean = product.Mean;
                    variance = product.Variance;
                }
                for (int i = 0; i < length; i++)
                {
                    if (array[i].Precision <= precisionThreshold)
                    {
                        result[i] = Gaussian.FromMeanAndVariance(mean[i], variance[i, i]) / array[i];
                    }
                    else
                    {
                        double alpha = meanTimesPrecision[i];
                        double beta = precision[i, i];
                        if (double.IsNaN(alpha)) throw new Exception("alpha is NaN");
                        if (double.IsNaN(beta)) throw new Exception("beta is NaN");
                        result[i] = GaussianOp.GaussianFromAlphaBeta(array[i], alpha, beta, false);
                        if (double.IsNaN(result[i].MeanTimesPrecision)) throw new Exception("result is NaN");
                        if (result[i].Precision < 0) throw new Exception("result is improper");
                    }
                }
            }
            return result;
        }

        /// <summary>VMP message to <c>array</c>.</summary>
        /// <param name="vector">Incoming message from <c>vector</c>. Must be a proper distribution. If any element is uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>array</c> as the random arguments are varied. The formula is <c>proj[sum_(vector) p(vector) factor(array,vector)]</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="vector" /> is not a proper distribution.</exception>
        /// <typeparam name="GaussianList">The type of the resulting array.</typeparam>
        public static GaussianList ArrayAverageLogarithm<GaussianList>([SkipIfUniform] VectorGaussian vector, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            if (result.Count != vector.Dimension)
                throw new ArgumentException("vector.Dimension (" + vector.Dimension + ") != result.Count (" + result.Count + ")");
            int length = result.Count;
            Vector mean = Vector.Zero(length);
            PositiveDefiniteMatrix variance = new PositiveDefiniteMatrix(length, length);
            vector.GetMeanAndVariance(mean, variance);
            for (int i = 0; i < length; i++)
            {
                result[i] = Gaussian.FromMeanAndVariance(mean[i], variance[i, i]);
            }
            return result;
        }
    }
}
