/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using System.Collections.Generic;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

    // using GaussianArray = MicrosoftResearch.Infer.Distributions.DistributionStructArray<MicrosoftResearch.Infer.Distributions.Gaussian, double>;

       /// <summary>Provides outgoing messages for <see cref="Vector.InnerProduct(Vector, Vector)" />, given random arguments to the function.</summary>
    [FactorMethod(typeof(Factor), "InnerProduct", typeof(double[]), typeof(double[]))]
    [Quality(QualityBand.Experimental)]
    public static class InnerProductArrayOp
    {
        //-- VMP -------------------------------------------------------------------------------------------------------------

        [Skip]
        public static double AverageLogFactor()
        {
            return 0.0;
        }

        public static Gaussian InnerProductAverageLogarithm(
            [SkipIfUniform] IList<Gaussian> A, [SkipIfUniform] IList<Gaussian> B, Gaussian result)
        {
            if (result == null)
                result = new Gaussian();
            // x[i,j] = sum_k a[i,k]*b[k,j]
            // E(x[i,j]) = sum_k E(a[i,k])*E(b[k,j])
            // var(x[i,j]) = sum_k E(a[i,k])^2*var(b[k,j]) + var(a[i,k])*E(b[k,j])^2 + var(a[i,k])*var(b[k,j])
            double mean = 0;
            double var = 0;
            for (int k = 0; k < A.Count; k++)
            {
                double Am, Av, Bm, Bv;
                A[k].GetMeanAndVariance(out Am, out Av);
                B[k].GetMeanAndVariance(out Bm, out Bv);
                mean += Am * Bm;
                if (double.IsPositiveInfinity(Av) || double.IsPositiveInfinity(Bv))
                    var = double.PositiveInfinity;
                else
                    var += Av * Bm * Bm + Bv * Am * Am + Av * Bv;
            }
            Gaussian r = result;
            r.SetMeanAndVariance(mean, var);
            result = r;
            return result;
        }

        [Skip]
        public static Gaussian InnerProductAverageLogarithmInit([IgnoreDependency] IList<Gaussian> A, [IgnoreDependency] IList<Gaussian> B)
        {
            return new Gaussian();
        }

        [Skip]
        public static Gaussian InnerProductAverageLogarithmInit([IgnoreDependency] double[] A, [IgnoreDependency] IList<Gaussian> B)
        {
            return new Gaussian();
        }

        [Skip]
        public static Gaussian InnerProductAverageLogarithmInit([IgnoreDependency] IList<Gaussian> A, [IgnoreDependency] double[] B)
        {
            return new Gaussian();
        }

        public static Gaussian InnerProductAverageLogarithm(double[] A, [SkipIfUniform] IList<Gaussian> B, Gaussian result)
        {
            return InnerProductAverageConditional(A, B, result);
        }

        public static Gaussian InnerProductAverageLogarithm([SkipIfUniform] IList<Gaussian> A, double[] B, Gaussian result)
        {
            return InnerProductAverageConditional(A, B, result);
        }

        public static GaussianList AAverageLogarithm<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, [Stochastic] IList<Gaussian> A, [SkipIfUniform] IList<Gaussian> B, GaussianList to_A)
            where GaussianList : IList<Gaussian> 
        {
            GaussianList result = to_A;
            //if (result == null)
            //{
            //    result = new GaussianList(A.Count);
            //    for (int k = 0; k < A.Count; k++)
            //        result[k] = new Gaussian();
            //}

            double ipMean = 0;
            for (int k = 0; k < A.Count; k++)
                ipMean += A[k].GetMean() * B[k].GetMean();

            double aMean,
                   bMean, bVariance;

            Gaussian result_old = new Gaussian();
            Gaussian cc;

            for (int k = 0; k < A.Count; k++)
            {
                aMean = A[k].GetMean();
                B[k].GetMeanAndVariance(out bMean, out bVariance);


                //result_old.SetNatural(result[k].MeanTimesPrecision, result[k].Precision);
                result_old.SetTo(result[k]);
                cc = result[k];
                
                if (!innerProduct.IsPointMass)
                    cc.SetNatural( bMean * (innerProduct.MeanTimesPrecision - innerProduct.Precision * (ipMean - aMean * bMean)), innerProduct.Precision * (bVariance + bMean * bMean));
                else
                    cc.SetTo(Gaussian.PointMass(bMean * (innerProduct.Point - (ipMean - aMean * bMean)) / (bVariance + bMean * bMean)));                    

                result[k] =  cc;

                // book-keeping for ipMean so that we have a sequential scheduling/coordinate descent for the elements of A 
                Gaussian newPost = A[k] * result[k] / result_old;           // update the global marginal
                ipMean  = ipMean - (aMean - newPost.GetMean()) * bMean;     // update the dependent quantities
            }

            return result;
        }

        public static GaussianList AAverageLogarithm<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, [Proper, Stochastic] IList<Gaussian> A, double[] B, GaussianList to_A)
            where GaussianList : IList<Gaussian> 
        {

            GaussianList result = to_A;
            //if (result == null)
            //{
            //    result = new DistributionStructArray<Gaussian, double>(A.Count);
            //    for (int k = 0; k < A.Count; k++)
            //        result[k] = new Gaussian();
            //}
            
            double ipMean = 0;
            for (int k = 0; k < A.Count; k++)
                ipMean += A[k].GetMean() * B[k];

            double aMean;

            Gaussian result_old = new Gaussian();
            Gaussian cc;

            for (int k = 0; k < A.Count; k++)
            {
                aMean = A[k].GetMean();

                //result_old.SetNatural(result[k].MeanTimesPrecision, result[k].Precision);
                result_old.SetTo(result[k]);
                cc = result[k];

                if (!innerProduct.IsPointMass)
                    cc.SetNatural(B[k] * (innerProduct.MeanTimesPrecision - innerProduct.Precision * (ipMean - aMean * B[k])), innerProduct.Precision * (B[k] * B[k]));
                else
                    cc.SetTo(Gaussian.PointMass(B[k] * (innerProduct.Point - (ipMean - aMean * B[k])) /  (B[k] * B[k])));    

                result[k] = cc;

                // book-keeping for ipMean so that we have a sequential scheduling/coordinate descent for the elements of A 
                Gaussian newPost = A[k] * result[k] / result_old;          // update the global marginal
                ipMean = ipMean - (aMean - newPost.GetMean()) * B[k];      // update the dependent quantities
            }

            return result;
        }

        public static GaussianList BAverageLogarithm<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, [SkipIfUniform] IList<Gaussian> A, [Proper, Stochastic] IList<Gaussian> B, GaussianList to_B)
            where GaussianList : IList<Gaussian> 
        {
            return AAverageLogarithm(innerProduct, B, A, to_B);
        }

        public static GaussianList BAverageLogarithm<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, double[] A, [Proper, Stochastic] IList<Gaussian> B, GaussianList to_B)
            where GaussianList : IList<Gaussian> 
        {
            return AAverageLogarithm(innerProduct, B, A, to_B);
        }

        private const string NotSupportedMessage = "Variational Message Passing does not support a InnerProduct factor with fixed output.";

        [NotSupported(InnerProductArrayOp.NotSupportedMessage)]
        public static GaussianList AAverageLogarithm<GaussianList>(double innerProduct)
               where GaussianList : IList<Gaussian> 
        {
            throw new NotSupportedException(InnerProductArrayOp.NotSupportedMessage);
        }

        [NotSupported(InnerProductArrayOp.NotSupportedMessage)]
        public static IList<Gaussian> BAverageLogarithm(double innerProduct)
        {
            throw new NotSupportedException(InnerProductArrayOp.NotSupportedMessage);
        }

        // AverageConditional ------------------------------------------------------------------------------------------------------------------------

        private const string LowRankNotSupportedMessage = "A InnerProduct factor with fixed output is not yet implemented for Expectation Propagation.";

        private const string BothRandomNotSupportedMessage =
            "A InnerProduct factor between two Gaussian arrays is not yet implemented for Expectation Propagation.  Try using Variational Message Passing.";

        [NotSupported(InnerProductArrayOp.BothRandomNotSupportedMessage)]
        public static Gaussian InnerProductAverageConditional([SkipIfUniform] IList<Gaussian> A, [SkipIfUniform] IList<Gaussian> B, IList<Gaussian> result)
        {
            throw new NotSupportedException(InnerProductArrayOp.BothRandomNotSupportedMessage);
        }

        [NotSupported(InnerProductArrayOp.BothRandomNotSupportedMessage)]
        public static GaussianList AAverageConditional<GaussianList>(Gaussian innerProduct, [SkipIfUniform] IList<Gaussian> B, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            throw new NotSupportedException(InnerProductArrayOp.BothRandomNotSupportedMessage);
        }

        [NotSupported(InnerProductArrayOp.BothRandomNotSupportedMessage)]
        public static GaussianList BAverageConditional<GaussianList>(Gaussian innerProduct, [SkipIfUniform] IList<Gaussian> A, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            throw new NotSupportedException(InnerProductArrayOp.BothRandomNotSupportedMessage);
        }

        public static double LogAverageFactor(double innerProduct, double[] A, double[] B)
        {

            double sum = 0;
            for (int k = 0; k < A.GetLength(0); k++)
            {
                sum += A[k] * B[k];
            }
            if (innerProduct != sum)
                return Double.NegativeInfinity;
            // return sum

            return 0.0;
        }

        public static double LogEvidenceRatio(double innerProduct, double[] A, double[] B)
        {
            return LogAverageFactor(innerProduct, A, B);
        }

        public static double AverageLogFactor(double innerProduct, double[] A, double[] B)
        {
            return LogAverageFactor(innerProduct, A, B);
        }

        public static double LogAverageFactor(Gaussian innerProduct, double[] A, IList<Gaussian> B)
        {
            Gaussian dummy = new Gaussian();
            Gaussian to_innerProduct = InnerProductAverageConditional(A, B, dummy);
            return to_innerProduct.GetLogAverageOf(innerProduct);
        }

        public static double LogAverageFactor(double innerProduct, double[] A, IList<Gaussian> B)
        {
            Gaussian dummy = new Gaussian();
            Gaussian to_innerProduct = InnerProductAverageConditional(A, B, dummy);
            return to_innerProduct.GetLogProb(innerProduct);
        }

        public static double LogAverageFactor(Gaussian innerProduct, IList<Gaussian> A, double[] B)
        {
            Gaussian dummy = new Gaussian();
            Gaussian to_innerProduct = InnerProductAverageConditional(A, B, dummy);
            return to_innerProduct.GetLogAverageOf(innerProduct);
        }

        public static double LogAverageFactor(double innerProduct, IList<Gaussian> A, double[] B)
        {
            Gaussian dummy = new Gaussian();
            Gaussian to_innerProduct = InnerProductAverageConditional(A, B, dummy);
            return to_innerProduct.GetLogProb(innerProduct);
        }

        [Skip]
        public static double LogEvidenceRatio(Gaussian innerProduct, double[] A, IList<Gaussian> B)
        {
            return 0.0;
        }

        [Skip]
        public static double LogEvidenceRatio(Gaussian innerProduct, IList<Gaussian> A, double[] B)
        {
            return 0.0;
        }

        public static double LogEvidenceRatio(double innerProduct, double[] A, IList<Gaussian> B)
        {
            return LogAverageFactor(innerProduct, A, B);
        }

        public static double LogEvidenceRatio(double innerProduct, IList<Gaussian> A, double[] B)
        {
            return LogAverageFactor(innerProduct, A, B);
        }

        
        public static Gaussian InnerProductAverageConditional(double[] A, [SkipIfUniform] IList<Gaussian> B, Gaussian result)
        {

            if (result == null)
                result = new Gaussian();

            double xMean = 0, xVariance = 0,
                   bMean, bVariance;

            for (int k = 0; k < A.Length; k++)
            {
                B[k].GetMeanAndVariance(out bMean, out bVariance);
                xMean += A[k] * bMean;
                xVariance += A[k] * A[k] * bVariance;
            }

            result.SetMeanAndVariance(xMean, xVariance);
            // Console.WriteLine("{0}, {1}", xMean, xVariance);
                        
            return result;
        }

        public static Gaussian InnerProductAverageConditional([SkipIfUniform] IList<Gaussian> A, double[] B, Gaussian result)
        {
            return InnerProductAverageConditional(B, A, result);
        }

        public static GaussianList AAverageConditional<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, [SkipIfUniform] IList<Gaussian> A, double[] B, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            //if (result == null)
            //{
            //    result = new GaussianList(A.Count);
            //    for (int k = 0; k < A.Count; k++)
            //        result[k] = new Gaussian();
            //}

            double xMean, xVariance;
            innerProduct.GetMeanAndVariance(out xMean, out xVariance);

            double aMean, aVariance;
            double qfBAB = xVariance, ipAB = 0;

            for (int k = 0; k < B.Length; k++)
            {
                A[k].GetMeanAndVariance(out aMean, out aVariance);
                ipAB  += B[k] * aMean;
                qfBAB += B[k] * B[k] * aVariance;
            }

            double scale = (xMean - ipAB)/qfBAB;

            Gaussian cc;
            for (int k = 0; k < B.Length; k++)
            {
                A[k].GetMeanAndVariance(out aMean, out aVariance);

                // for the moment leave error checking to others
                //  double aVarianceNew = aVariance - (B[k] * B[k] * aVariance * aVariance) / qfBAB;

                //cc = result[k];
                //cc.SetMeanAndVariance(aMean + s*B[k] * aVariance,
                //                             aVariance - (B[k] * B[k] * aVariance * aVariance) / qfBAB);
                //result[k] = cc;
                //result[k] = result[k] / A[k];

                // do some damped update
                Gaussian msg_old = new Gaussian(result[k]);
                Gaussian msg_new = new Gaussian(aMean + scale * B[k] * aVariance, aVariance - (B[k] * B[k] * aVariance * aVariance) / qfBAB);
                msg_new = msg_new / A[k];
                double deps = 0.62;
                msg_new.SetToPower(msg_new, deps);
                msg_old.SetToPower(msg_old, 1-deps);
                
                cc = result[k];
                cc = msg_new * msg_old;
                result[k] = cc;
            }

            //for (int k = 0; k < B.Length; k++)
            //{
            //    result[k] = result[k] / A[k];
            //}

            return result;
        }

        public static GaussianList BAverageConditional<GaussianList>(
            [SkipIfUniform] Gaussian innerProduct, double[] A, [SkipIfUniform] IList<Gaussian> B, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            return AAverageConditional(innerProduct, B, A, result);
        }

        [NotSupported(InnerProductArrayOp.LowRankNotSupportedMessage)]
        public static GaussianList AAverageConditional<GaussianList>(double innerProduct, double[] B, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            throw new NotImplementedException(InnerProductArrayOp.LowRankNotSupportedMessage);
        }

        [NotSupported(InnerProductArrayOp.LowRankNotSupportedMessage)]
        public static GaussianList BAverageConditional<GaussianList>(double innerProduct, double A, GaussianList result)
            where GaussianList : IList<Gaussian>
        {
            throw new NotImplementedException(InnerProductArrayOp.LowRankNotSupportedMessage);
        }
    }
}
