/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using MicrosoftResearch.Infer.Distributions;

    [FactorMethod(typeof(Factor), "IsGreaterThan", typeof(double), typeof(double))]
    [Quality(QualityBand.Preview)]
    public static class IsGreaterThanDoubleOp
    {
        // Beta ---------------------------------------------------------------------------------------

        public static Bernoulli IsGreaterThanAverageConditional([Proper] Beta a, double b)
        {
            if (a.IsPointMass)
                return Bernoulli.PointMass(a.Point > b);
            return new Bernoulli(1 - a.GetProbLessThan(b));
        }

        public static Bernoulli IsGreaterThanAverageConditional(double a, [Proper] Beta b)
        {
            if (b.IsPointMass)
                return Bernoulli.PointMass(a > b.Point);
            return new Bernoulli(b.GetProbLessThan(a));
        }

        public static Bernoulli IsGreaterThanAverageConditional([Proper] Beta a, [Proper] Beta b)
        {
            if (a.IsPointMass)
                return IsGreaterThanAverageConditional(a.Point, b);
            if (b.IsPointMass)
                return IsGreaterThanAverageConditional(a, b.Point);
            double aVariance = a.GetVariance();
            double bVariance = b.GetVariance();
            int sampleCount = 10000;
            double sum = 0;
            for (int sampleIndex = 0; sampleIndex < sampleCount; sampleIndex++)
            {
                if (aVariance < bVariance)
                {
                    double aSample = a.Sample();
                    sum += b.GetProbLessThan(aSample);
                }
                else
                {
                    double bSample = b.Sample();
                    sum += 1 - a.GetProbLessThan(bSample);
                }
            }
            double prob = sum / sampleCount;
            return new Bernoulli(prob);
        }

        public static Beta AAverageConditional([SkipIfUniform] Bernoulli isGreaterThan, Beta a, Beta b)
        {
            throw new NotImplementedException();
        }

        public static Beta BAverageConditional([SkipIfUniform] Bernoulli isGreaterThan, Beta a, Beta b)
        {
            throw new NotImplementedException();
        }

        // Gamma ---------------------------------------------------------------------------------------

        public static Bernoulli IsGreaterThanAverageConditional([Proper] Gamma a, double b)
        {
            if (a.IsPointMass)
                return Bernoulli.PointMass(a.Point > b);
            return new Bernoulli(1 - a.GetProbLessThan(b));
        }

        public static Bernoulli IsGreaterThanAverageConditional(double a, [Proper] Gamma b)
        {
            if (b.IsPointMass)
                return Bernoulli.PointMass(a > b.Point);
            return new Bernoulli(b.GetProbLessThan(a));
        }

        public static Bernoulli IsGreaterThanAverageConditional([Proper] Gamma a, [Proper] Gamma b)
        {
            if (a.IsPointMass)
                return IsGreaterThanAverageConditional(a.Point, b);
            if (b.IsPointMass)
                return IsGreaterThanAverageConditional(a, b.Point);
            if (a.Rate == b.Rate)
            {
                // If a and b have the same Rate, then c = b/(b+a) is Beta(b.Shape, a.Shape).
                // (a > b) is equivalent to (c < 0.5).
                Beta c = new Beta(b.Shape, a.Shape);
                return new Bernoulli(c.GetProbLessThan(0.5));
            }
            double aVariance = a.GetVariance();
            double bVariance = b.GetVariance();
            int sampleCount = 10000;
            double sum = 0;
            for (int sampleIndex = 0; sampleIndex < sampleCount; sampleIndex++)
            {
                if (aVariance < bVariance)
                {
                    double aSample = a.Sample();
                    sum += b.GetProbLessThan(aSample);
                }
                else
                {
                    double bSample = b.Sample();
                    sum += 1 - a.GetProbLessThan(bSample);
                }
            }
            double prob = sum / sampleCount;
            return new Bernoulli(prob);
        }

        public static Gamma AAverageConditional([SkipIfUniform] Bernoulli isGreaterThan, Gamma a, Gamma b)
        {
            throw new NotImplementedException();
        }

        public static Gamma BAverageConditional([SkipIfUniform] Bernoulli isGreaterThan, Gamma a, Gamma b)
        {
            throw new NotImplementedException();
        }

        public static TruncatedGamma AAverageConditional(bool isGreaterThan, double b)
        {
            if (isGreaterThan) return new TruncatedGamma(Gamma.Uniform(), b, double.PositiveInfinity);
            else return new TruncatedGamma(Gamma.Uniform(), 0, b);
        }

        public static TruncatedGamma BAverageConditional(bool isGreaterThan, double a)
        {
            return AAverageConditional(!isGreaterThan, a);
        }
    }
}
