/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

    /// <summary>Provides outgoing messages for <see cref="Factor.Product(Matrix, Vector)" />, given random arguments to the function.</summary>
    [FactorMethod(typeof(Factor), "Product", typeof(Matrix), typeof(Vector))]
    [FactorMethod(typeof(Factor), "Product", typeof(double[,]), typeof(Vector))]
    [Buffers("BMean", "BVariance")]
    [Quality(QualityBand.Stable)]
    public static class MatrixVectorProductOp
    {
        public static bool UseAccurateMethod = true;

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(product,a,b))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(Vector product, Matrix A, Vector B)
        {
            return product.Equals(Factor.Product(A, B)) ? 0.0 : Double.NegativeInfinity;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(product,a,b))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(Vector product, Matrix A, Vector B)
        {
            return LogAverageFactor(product, A, B);
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(product,a,b))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        public static double AverageLogFactor(Vector product, Matrix A, Vector B)
        {
            return LogAverageFactor(product, A, B);
        }

        /// <summary>Initialize the buffer <c>BVariance</c>.</summary>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <returns>Initial value of buffer <c>BVariance</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static PositiveDefiniteMatrix BVarianceInit([IgnoreDependency] VectorGaussian B)
        {
            return new PositiveDefiniteMatrix(B.Dimension, B.Dimension);
        }

        /// <summary>Update the buffer <c>BVariance</c>.</summary>
        /// <param name="B">Incoming message from <c>b</c>. Must be a proper distribution. If any element is uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="B" /> is not a proper distribution.</exception>
        [Fresh]
        public static PositiveDefiniteMatrix BVariance([Proper] VectorGaussian B, PositiveDefiniteMatrix result)
        {
            return B.GetVariance(result);
        }

        /// <summary>Initialize the buffer <c>BMean</c>.</summary>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <returns>Initial value of buffer <c>BMean</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static Vector BMeanInit([IgnoreDependency] VectorGaussian B)
        {
            return Vector.Zero(B.Dimension);
        }

        /// <summary>Update the buffer <c>BMean</c>.</summary>
        /// <param name="B">Incoming message from <c>b</c>. Must be a proper distribution. If any element is uniform, the result will be uniform.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="B" /> is not a proper distribution.</exception>
        [Fresh]
        public static Vector BMean([Proper] VectorGaussian B, PositiveDefiniteMatrix BVariance, Vector result)
        {
            return B.GetMean(result, BVariance);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <param name="to_product">Outgoing message to <c>product</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(product) p(product) factor(product,a,b))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(VectorGaussian product, [Fresh] VectorGaussian to_product)
        {
            return to_product.GetLogAverageOf(product);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <param name="BMean">Buffer <c>BMean</c>.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(b) p(b) factor(product,a,b))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(Vector product, Matrix A, VectorGaussian B, Vector BMean, PositiveDefiniteMatrix BVariance)
        {
            VectorGaussian toProduct = ProductAverageConditional(A, BMean, BVariance, new VectorGaussian(A.Rows));
            return toProduct.GetLogProb(product);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(product,b) p(product,b) factor(product,a,b) / sum_product p(product) messageTo(product))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        [Skip]
        public static double LogEvidenceRatio(VectorGaussian product, Matrix A, VectorGaussian B)
        {
            return 0.0;
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(VectorGaussian product, Matrix A, VectorGaussian B)
        {
            return 0.0;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <param name="BMean">Buffer <c>BMean</c>.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(b) p(b) factor(product,a,b))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(Vector product, Matrix A, VectorGaussian B, Vector BMean, PositiveDefiniteMatrix BVariance)
        {
            return LogAverageFactor(product, A, B, BMean, BVariance);
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>.</param>
        /// <param name="BMean">Buffer <c>BMean</c>.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        public static double AverageLogFactor(Vector product, Matrix A, VectorGaussian B, Vector BMean, PositiveDefiniteMatrix BVariance)
        {
            return LogAverageFactor(product, A, B, BMean, BVariance);
        }

        private static void GetProductMoments(Matrix A, Vector BMean, PositiveDefiniteMatrix BVariance, Vector mean, PositiveDefiniteMatrix variance)
        {
            // P.mean = A*B.mean
            // P.var = A*B.var*A'
            mean.SetToProduct(A, BMean);
            if (UseAccurateMethod)
            {
                int dim = BVariance.Rows;
                LowerTriangularMatrix cholesky = new LowerTriangularMatrix(dim, dim);
                cholesky.SetToCholesky(BVariance);
                Matrix AL = A * cholesky;
                variance.SetToOuter(AL);
            }
            else
            {
                variance.SetToProduct(A, (A * BVariance).Transpose());
                variance.Symmetrize();
            }
        }

        /// <summary>EP message to <c>product</c>.</summary>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="BMean">Buffer <c>BMean</c>.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>product</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussian ProductAverageConditional(Matrix A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussian result)
        {
            Vector mean = Vector.Zero(result.Dimension);
            PositiveDefiniteMatrix variance = result.Precision;
            GetProductMoments(A, BMean, BVariance, mean, variance);
            result.SetMeanAndVariance(mean, variance);
            return result;
        }

        public static VectorGaussianMoments ProductAverageConditional(Matrix A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussianMoments result)
        {
            GetProductMoments(A, BMean, BVariance, result.Mean, result.Variance);
            return result;
        }

        public static VectorGaussian ProductAverageConditional(double[,] A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussian result)
        {
            return ProductAverageConditional(new Matrix(A), BMean, BVariance, result);
        }

        public static VectorGaussianMoments ProductAverageConditional(double[,] A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussianMoments result)
        {
            return ProductAverageConditional(new Matrix(A), BMean, BVariance, result);
        }

        public static VectorGaussian ProductAverageConditional(DistributionArray2D<Gaussian, double> A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussian result)
        {
            if (!A.IsPointMass) throw new ArgumentException("A is not a point mass");
            return ProductAverageConditional(new Matrix(A.Point), BMean, BVariance, result);
        }

        public static VectorGaussianMoments ProductAverageConditional(DistributionArray2D<Gaussian, double> A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussianMoments result)
        {
            if (!A.IsPointMass) throw new ArgumentException("A is not a point mass");
            return ProductAverageConditional(new Matrix(A.Point), BMean, BVariance, result);
        }

        /// <summary />
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static VectorGaussian ProductAverageConditionalInit([IgnoreDependency] Matrix A)
        {
            return new VectorGaussian(A.Rows);
        }

        /// <summary />
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static VectorGaussian ProductAverageLogarithmInit([IgnoreDependency] Matrix A)
        {
            return new VectorGaussian(A.Rows);
        }

        /// <summary>VMP message to <c>product</c>.</summary>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="BMean">Buffer <c>BMean</c>.</param>
        /// <param name="BVariance">Buffer <c>BVariance</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>product</c> conditioned on the given values.</para>
        /// </remarks>
        public static VectorGaussian ProductAverageLogarithm(Matrix A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussian result)
        {
            return ProductAverageConditional(A, BMean, BVariance, result);
        }

        public static VectorGaussianMoments ProductAverageLogarithm(Matrix A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussianMoments result)
        {
            return ProductAverageConditional(A, BMean, BVariance, result);
        }

        public static VectorGaussian ProductAverageLogarithm(double[,] A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussian result)
        {
            return ProductAverageConditional(A, BMean, BVariance, result);
        }

        public static VectorGaussianMoments ProductAverageLogarithm(double[,] A, Vector BMean, PositiveDefiniteMatrix BVariance, VectorGaussianMoments result)
        {
            return ProductAverageConditional(A, BMean, BVariance, result);
        }

        /// <summary>EP message to <c>b</c>.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>b</c> as the random arguments are varied. The formula is <c>proj[p(b) sum_(product) p(product) factor(product,a,b)]/p(b)</c>.</para>
        /// </remarks>
        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussian product, Matrix A, VectorGaussian result)
        {
            if (product.IsPointMass)
                return BAverageConditional(product.Point, A, result);
            //   (p.mean - A*B)'*p.prec*(p.mean - A*B)
            // = B'*(A'*p.prec*A)*B - B'*A'*p.prec*p.mean - ...
            // B.prec = A'*p.prec*A
            // B.precTimesMean = A'*p.precTimesMean
            if (UseAccurateMethod)
            {
                // this method is slower but more numerically accurate
                // L*L' = p.prec
                int dim = product.Precision.Cols;
                LowerTriangularMatrix L = new LowerTriangularMatrix(dim, dim);
                L.SetToCholesky(product.Precision);                
                Matrix At = A.Transpose();
                Matrix temp = At * L;
                result.Precision.SetToOuter(temp);
            }
            else
            {
                Matrix temp = (product.Precision * A).Transpose();
                result.Precision.SetToProduct(temp, A);
                result.Precision.Symmetrize();
            }
            result.MeanTimesPrecision.SetToProduct(product.MeanTimesPrecision, A);
            return result;
        }

        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussianMoments product, Matrix A, VectorGaussian result)
        {
            if (product.IsPointMass)
                return BAverageConditional(product.Point, A, result);
            VectorGaussian productNatural = new VectorGaussian(product.Mean, product.Variance);
            return BAverageConditional(productNatural, A, result);
        }

        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussian product, double[,] A, VectorGaussian result)
        {
            return BAverageConditional(product, new Matrix(A), result);
        }

        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussianMoments product, double[,] A, VectorGaussian result)
        {
            return BAverageConditional(product, new Matrix(A), result);
        }

        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussian product, DistributionArray2D<Gaussian,double> A, VectorGaussian result)
        {
            if (!A.IsPointMass) throw new ArgumentException("A is not a point mass");
            return BAverageConditional(product, new Matrix(A.Point), result);
        }

        public static VectorGaussian BAverageConditional([SkipIfUniform] VectorGaussianMoments product, DistributionArray2D<Gaussian, double> A, VectorGaussian result)
        {
            if (!A.IsPointMass) throw new ArgumentException("A is not a point mass");
            return BAverageConditional(product, new Matrix(A.Point), result);
        }

        /// <summary>VMP message to <c>b</c>.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>b</c> with <c>product</c> integrated out. The formula is <c>sum_product p(product) factor(product,a,b)</c>.</para>
        /// </remarks>
        public static VectorGaussian BAverageLogarithm([SkipIfUniform] VectorGaussian product, Matrix A, VectorGaussian result)
        {
            return BAverageConditional(product, A, result);
        }

        public static VectorGaussian BAverageLogarithm([SkipIfUniform] VectorGaussianMoments product, Matrix A, VectorGaussian result)
        {
            return BAverageConditional(product, A, result);
        }

        private const string LowRankNotSupportedMessage = "A matrix-vector product with fixed output is not yet implemented.";

        /// <summary>EP message to <c>b</c>.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>b</c> conditioned on the given values.</para>
        /// </remarks>
        [NotSupported(MatrixVectorProductOp.LowRankNotSupportedMessage)]
        public static VectorGaussian BAverageConditional(Vector product, Matrix A, VectorGaussian result)
        {
            throw new NotSupportedException(MatrixVectorProductOp.LowRankNotSupportedMessage);
        }

        /// <summary>VMP message to <c>b</c>.</summary>
        /// <param name="product">Constant value for <c>product</c>.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>b</c> conditioned on the given values.</para>
        /// </remarks>
        [NotSupported(MatrixVectorProductOp.LowRankNotSupportedMessage)]
        public static VectorGaussian BAverageLogarithm(Vector product, Matrix A, VectorGaussian result)
        {
            throw new NotSupportedException(MatrixVectorProductOp.LowRankNotSupportedMessage);
        }

        public static DistributionStructArray2D<Gaussian, double> AAverageConditional([SkipIfUniform] VectorGaussian product, DistributionArray2D<Gaussian, double> A, Vector BMean, PositiveDefiniteMatrix BVariance, DistributionStructArray2D<Gaussian, double> result)
        {
            if (product.IsUniform())
            {
                result.SetToUniform();
                return result;
            }
            if (!A.IsPointMass) throw new ArgumentException("A is not a point mass");
            // logZ = log N(mProduct; A*BMean, vProduct + A*BVariance*A')
            //      = -0.5 (mProduct - A*BMean)' inv(vProduct + A*BVariance*A') (mProduct - A*BMean) - 0.5 logdet(vProduct + A*BVariance*A')
            //      = -0.5 (mProduct - A*BMean)' pPrec inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (mProduct - A*BMean) 
            //        - 0.5 logdet(pProduct + pProduct*A*BVariance*A'*pProduct) + logdet(pProduct)
            // dlogZ   = 0.5 (dA*BMean)' pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (mProduct - A*BMean) 
            //         +0.5 (mProduct - A*BMean)' pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (dA*BMean) 
            //         +0.5 (mProduct - A*BMean)' pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) (pProduct*dA*BVariance*A'*pProduct + pProduct*A*BVariance*dA'*pProduct) inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (mProduct - A*BMean) 
            //         - 0.5 tr(inv(pProduct + pProduct*A*BVariance*A'*pProduct) (pProduct*dA*BVariance*A'*pProduct + pProduct*A*BVariance*dA'*pProduct))
            // dlogZ/dA = pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (mProduct - A*BMean) BMean'
            //          + pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct (mProduct - A*BMean) (mProduct - A*BMean)' pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct*A*BVariance
            //          - pProduct inv(pProduct + pProduct*A*BVariance*A'*pProduct) pProduct A*BVariance
            var Amatrix = new Matrix(A.Point);
            var pProductA = product.Precision * Amatrix;
            var pProductABV = pProductA * BVariance;
            PositiveDefiniteMatrix prec = new PositiveDefiniteMatrix(product.Dimension, product.Dimension);
            prec.SetToSum(product.Precision, pProductABV * pProductA.Transpose());
            // pProductA is now free
            for (int i = 0; i < prec.Rows; i++)
            {
                if (prec[i, i] == 0) prec[i, i] = 1;
            }
            var v = prec.Inverse();
            var ABM = Amatrix * BMean;
            var pProductABM = product.Precision * ABM;
            var diff = pProductABM;
            diff.SetToDifference(product.MeanTimesPrecision, pProductABM);
            // ABM is now free
            var pProductV = product.Precision * v;
            var pProductVdiff = ABM;
            pProductVdiff.SetToProduct(pProductV, diff);
            var Vdiff = v * diff;
            pProductV.Scale(-1);
            pProductV.SetToSumWithOuter(pProductV, 1, pProductVdiff, Vdiff);
            Matrix dlogZ = pProductA;
            dlogZ.SetToProduct(pProductV, pProductABV);
            dlogZ.SetToSumWithOuter(dlogZ, 1, pProductVdiff, BMean);
            int rows = A.GetLength(0);
            int cols = A.GetLength(1);
            for (int i = 0; i < rows; i++)
            {
                for (int j = 0; j < cols; j++)
                {
                    double dlogp = dlogZ[i,j];
                    // for now, we don't compute the second derivative.
                    double ddlogp = -1;
                    result[i, j] = Gaussian.FromDerivatives(A[i, j].Point, dlogp, ddlogp, false);
                }
            }
            return result;
        }
    }
}
