/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

       /// <summary>Provides outgoing messages for <see cref="Math.Max(double, double)" />, given random arguments to the function.</summary>
    [FactorMethod(new string[] { "max", "a", "b" }, typeof(Math), "Max", typeof(double), typeof(double))]
    [Quality(QualityBand.Stable)]
    public static class MaxGaussianOp
    {
        /// <summary>
        /// Static flag to force a proper distribution
        /// </summary>
        public static bool ForceProper = true;

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(max,val1,val2))</c>.</para>
        /// </remarks>
        public static double LogAverageFactor(double max, double a, double b)
        {
            return (max == Math.Max(a, b)) ? 0.0 : Double.NegativeInfinity;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(max,val1,val2))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        public static double LogEvidenceRatio(double max, double a, double b)
        {
            return LogAverageFactor(max, a, b);
        }

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(factor(max,val1,val2))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        public static double AverageLogFactor(double max, double a, double b)
        {
            return LogAverageFactor(max, a, b);
        }

        // logz is the log-expectation of the factor value.
        // logw1 is the log-probability that max==a
        // exp(logw1) = N(mx;m1,vx+v1) phi((mx1 - m2)/sqrt(vx1+v2))
        // alpha1 is the correction to the mean of a due to b given that max==a
        // alpha1 = N(mx1;m2,vx1+v2)/phi
        // vx1 is the variance of a given that max==a (ignoring b)
        // mx1 is the mean of a given that max==a (ignoring b)
        internal static void ComputeStats(Gaussian max, Gaussian a, Gaussian b, out double logz,
                                          out double logw1, out double alpha1, out double vx1, out double mx1,
                                          out double logw2, out double alpha2, out double vx2, out double mx2)
        {
            double m1, v1, m2, v2;
            a.GetMeanAndVariance(out m1, out v1);
            b.GetMeanAndVariance(out m2, out v2);
            if (max.IsPointMass)
            {
                vx1 = 0.0;
                mx1 = max.Point;
                vx2 = 0.0;
                mx2 = max.Point;
                if (b.IsPointMass)
                {
                    if (b.Point > max.Point)
                        throw new AllZeroException();
                    else if (b.Point == max.Point)
                    {
                        // the factor reduces to the constraint (max.Point >= a)
                        if (v1 == 0)
                        {
                            if (m1 > max.Point)
                                throw new AllZeroException();
                            if (m1 == max.Point)
                            {
                                logw1 = -MMath.Ln2;
                                logw2 = -MMath.Ln2;
                                logz = 0;
                                alpha1 = 0;
                                alpha2 = 0;
                                return;
                            }
                            else
                            {
                                logw2 = 0;
                            }
                        }
                        else if(double.IsPositiveInfinity(v1))
                        {
                            logw1 = double.NegativeInfinity;
                            logw2 = -MMath.Ln2;
                            logz = logw2;
                            alpha1 = 0;
                            alpha2 = 0;
                            return;
                        }
                        else
                        {
                            logw2 = MMath.NormalCdfLn((max.Point - m1) / Math.Sqrt(v1));
                        }
                        logw1 = double.NegativeInfinity;
                        logz = logw2;
                        alpha1 = 0;
                        alpha2 = Math.Exp(Gaussian.GetLogProb(max.Point, m1, v1) - logw2);
                        return;
                    }
                    else // b.Point < max.Point
                    {
                        // the factor reduces to the constraint (a == max.Point)
                        logw1 = a.GetLogProb(max.Point);
                        logw2 = double.NegativeInfinity;
                        logz = logw1;
                        alpha1 = 0;
                        alpha2 = 0;
                        return;
                    }
                }
                else if (a.IsPointMass) // !b.IsPointMass
                {
                    if (a.Point > max.Point)
                        throw new AllZeroException();
                    else if (a.Point == max.Point)
                    {
                        // the factor reduces to the constraint (max.Point > b)
                        if (double.IsPositiveInfinity(v2))
                        {
                            logw2 = double.NegativeInfinity;
                            logw1 = -MMath.Ln2;
                            logz = logw1;
                            alpha1 = 0;
                            alpha2 = 0;
                            return;
                        }
                        else
                        {
                            logw1 = MMath.NormalCdfLn((max.Point - m2) / Math.Sqrt(v2));
                            logw2 = double.NegativeInfinity;
                            logz = logw1;
                            alpha2 = 0;
                            alpha1 = Math.Exp(Gaussian.GetLogProb(max.Point, m2, v2) - logw1);
                            return;
                        }
                    }
                    else // a.Point < max.Point
                    {                        
                        // the factor reduces to the constraint (b == max.Point)
                        logw2 = b.GetLogProb(max.Point);
                        logw1 = double.NegativeInfinity;
                        logz = logw2;
                        alpha1 = 0;
                        alpha2 = 0;
                        return;
                    }
                }
            }
            else // !max.IsPointMass
            {
                if (a.IsPointMass)
                {
                    vx1 = 0.0;
                    mx1 = a.Point;
                    if (b.IsPointMass)
                    {
                        vx2 = 0.0;
                        mx2 = b.Point;
                        if (a.Point > b.Point)
                        {
                            logw1 = max.GetLogAverageOf(a);
                            logw2 = double.NegativeInfinity;
                            logz = logw1;
                        }
                        else if (a.Point < b.Point)
                        {
                            logw2 = max.GetLogAverageOf(b);
                            logw1 = double.NegativeInfinity;
                            logz = logw2;
                        }
                        else // a.Point == b.Point
                        {
                            logw1 = -MMath.Ln2;
                            logw2 = -MMath.Ln2;
                            logz = 0;
                        }
                        alpha1 = 0;
                        alpha2 = 0;
                        return;
                    }
                }
                else // !a.IsPointMass
                {
                    vx1 = 1.0 / (max.Precision + a.Precision);
                    mx1 = vx1 * (max.MeanTimesPrecision + a.MeanTimesPrecision);
                }
                if (b.IsPointMass)
                {
                    vx2 = 0.0;
                    mx2 = b.Point;
                }
                else // !b.IsPointMass
                {
                    vx2 = 1.0 / (max.Precision + b.Precision);
                    mx2 = vx2 * (max.MeanTimesPrecision + b.MeanTimesPrecision);
                }
            }
            // !max.IsPointMass
            // !(a.IsPointMass && b.IsPointMass)
            logw1 = max.GetLogAverageOf(a);
            double logPhi1 = MMath.NormalCdfLn((mx1 - m2) / Math.Sqrt(vx1 + v2));
            logw1 += logPhi1;

            logw2 = max.GetLogAverageOf(b);
            double logPhi2 = MMath.NormalCdfLn((mx2 - m1) / Math.Sqrt(vx2 + v1));
            logw2 += logPhi2;

            logz = MMath.LogSumExp(logw1, logw2);

            if (double.IsPositiveInfinity(v2))
                alpha1 = 0;
            else
            {
                double logN1 = Gaussian.GetLogProb(mx1, m2, vx1 + v2);
                alpha1 = Math.Exp(logN1 - logPhi1);
            }
            if (double.IsPositiveInfinity(v1))
                alpha2 = 0;
            else
            {
                double logN2 = Gaussian.GetLogProb(mx2, m1, vx2 + v1);
                alpha2 = Math.Exp(logN2 - logPhi2);
            }
            //double z2 = (mx2 - m1) / Math.Sqrt(vx2 + v1);
            //alpha2 = 1 / Math.Sqrt(vx2 + v1) / MMath.NormalCdfRatio(z2);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val1,val2) p(max,val1,val2) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogAverageFactor([SkipIfUniform] Gaussian max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            double logw1, a1, vx1, mx1;
            double logw2, a2, vx2, mx2;
            double logz;
            ComputeStats(max, a, b, out logz, out logw1, out a1, out vx1, out mx1,
                         out logw2, out a2, out vx2, out mx2);
            return logz;
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val2) p(max,val2) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogAverageFactor([SkipIfUniform] Gaussian max, double a, [Proper] Gaussian b)
        {
            return LogAverageFactor(max, Gaussian.PointMass(a), b);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val1) p(max,val1) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static double LogAverageFactor([SkipIfUniform] Gaussian max, [Proper] Gaussian a, double b)
        {
            return LogAverageFactor(max, a, Gaussian.PointMass(b));
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val1,val2) p(val1,val2) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogAverageFactor(double max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            return LogAverageFactor(Gaussian.PointMass(max), a, b);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val2) p(val2) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogAverageFactor(double max, double a, [Proper] Gaussian b)
        {
            return LogAverageFactor(Gaussian.PointMass(max), Gaussian.PointMass(a), b);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's average value across the given argument distributions.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val1) p(val1) factor(max,val1,val2))</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static double LogAverageFactor(double max, [Proper] Gaussian a, double b)
        {
            return LogAverageFactor(Gaussian.PointMass(max), a, Gaussian.PointMass(b));
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val1,val2) p(max,val1,val2) factor(max,val1,val2) / sum_max p(max) messageTo(max))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio([SkipIfUniform] Gaussian max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            Gaussian to_max = MaxAverageConditional(max, a, b);
            return LogAverageFactor(max, a, b)
                   - to_max.GetLogAverageOf(max);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val2) p(max,val2) factor(max,val1,val2) / sum_max p(max) messageTo(max))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio([SkipIfUniform] Gaussian max, double a, [Proper] Gaussian b)
        {
            Gaussian to_max = MaxAverageConditional(max, a, b);
            return LogAverageFactor(max, a, b)
                   - to_max.GetLogAverageOf(max);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(max,val1) p(max,val1) factor(max,val1,val2) / sum_max p(max) messageTo(max))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio([SkipIfUniform] Gaussian max, [Proper] Gaussian a, double b)
        {
            Gaussian to_max = MaxAverageConditional(max, a, b);
            return LogAverageFactor(max, a, b)
                   - to_max.GetLogAverageOf(max);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val1,val2) p(val1,val2) factor(max,val1,val2))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio(double max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            return LogAverageFactor(max, a, b);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val2) p(val2) factor(max,val1,val2))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio(double max, double a, [Proper] Gaussian b)
        {
            return LogAverageFactor(max, a, b);
        }

        /// <summary>Evidence message for EP.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(val1) p(val1) factor(max,val1,val2))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static double LogEvidenceRatio(double max, [Proper] Gaussian a, double b)
        {
            return LogAverageFactor(max, a, b);
        }

        /// <summary />
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        [Skip]
        public static Gaussian MaxAverageConditionalInit()
        {
            return Gaussian.Uniform();
        }

        /// <summary>EP message to <c>max</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>max</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>max</c> as the random arguments are varied. The formula is <c>proj[p(max) sum_(val2) p(val2) factor(max,val1,val2)]/p(max)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian MaxAverageConditional(Gaussian max, double a, [Proper] Gaussian b)
        {
            return MaxAverageConditional(max, Gaussian.PointMass(a), b);
        }

        /// <summary>EP message to <c>max</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>The outgoing EP message to the <c>max</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>max</c> as the random arguments are varied. The formula is <c>proj[p(max) sum_(val1) p(val1) factor(max,val1,val2)]/p(max)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static Gaussian MaxAverageConditional(Gaussian max, [Proper] Gaussian a, double b)
        {
            return MaxAverageConditional(max, a, Gaussian.PointMass(b));
        }

        /// <summary>EP message to <c>max</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>max</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>max</c> as the random arguments are varied. The formula is <c>proj[p(max) sum_(val1,val2) p(val1,val2) factor(max,val1,val2)]/p(max)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian MaxAverageConditional(Gaussian max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            if (max.IsPointMass)
                return Gaussian.Uniform();
            // the following code works correctly even if max is uniform or improper.
            if (!a.IsProper())
                throw new ImproperMessageException(a);
            if (!b.IsProper())
                throw new ImproperMessageException(b);
            double m1, v1, m2, v2;
            a.GetMeanAndVariance(out m1, out v1);
            b.GetMeanAndVariance(out m2, out v2);

            double logw1, alpha1, vx1, mx1;
            double logw2, alpha2, vx2, mx2;
            double logz;
            ComputeStats(max, a, b, out logz, out logw1, out alpha1, out vx1, out mx1,
                         out logw2, out alpha2, out vx2, out mx2);
            double w1 = Math.Exp(logw1 - logz);
            double w2 = Math.Exp(logw2 - logz);
            // the posterior is a mixture model with weights exp(logw1-logz), exp(logw2-logz) and distributions
            // N(x; mx1, vx1) phi((x - m2)/sqrt(v2)) / phi((mx1 - m2)/sqrt(vx1 + v2))
            // the moments of the posterior are computed via the moments of these two components.
            double mc1 = mx1 + alpha1 * vx1;
            double mc2 = mx2 + alpha2 * vx2;
            double m, v;
            if(w1 == 0)
            {
                // avoid dealing with infinities.
                m = mx2;
                v = vx2;
            }
            else if (w2 == 0)
            {
                // avoid dealing with infinities.
                m = mx1;
                v = vx1;
            }
            else
            {
                m = w1 * mc1 + w2 * mc2;
                double beta1;
                if (alpha1 == 0)
                    beta1 = 0;
                else
                {
                    double r1 = (mx1 - m2) / (vx1 + v2);
                    beta1 = alpha1 * (alpha1 + r1);
                }
                double vc1 = vx1 * (1 - vx1 * beta1);
                double beta2;
                if (alpha2 == 0)
                    beta2 = 0;
                else
                {
                    double r2 = (mx2 - m1) / (vx2 + v1);
                    beta2 = alpha2 * (alpha2 + r2);
                }
                double vc2 = vx2 * (1 - vx2 * beta2);
                double diff = mc1 - mc2;
                v = w1 * vc1 + w2 * vc2 + w1 * w2 * diff * diff;
            }
            Gaussian result = new Gaussian(m, v);
            result.SetToRatio(result, max, ForceProper);
            if (Double.IsNaN(result.Precision) || Double.IsNaN(result.MeanTimesPrecision))
                throw new InferRuntimeException($"result is NaN.  max={max}, a={a}, b={b}");
            return result;
        }

#if false
		public static void ComputeStats(Gaussian max, Gaussian a, Gaussian b, out double logz,
			out double logw1, out double logPhi1, out double logu1, out double vx1, out double mx1,
			out double logw2, out double logPhi2, out double logu2, out double vx2, out double mx2)
		{
			double mx, vx, ma, va, mb, vb;
			max.GetMeanAndVariance(out mx, out vx);
			a.GetMeanAndVariance(out ma, out va);
			b.GetMeanAndVariance(out mb, out vb);
			if (false) {
				vx1 = 1.0 / (1.0 / vx + 1.0 / va);
				mx1 = vx1 * (mx / vx + ma / va);
				vx2 = 1.0 / (1.0 / vx + 1.0 / vb);
				mx2 = vx2 * (mx / vx + mb / vb);
			} else {
				if(max.IsPointMass || a.IsPointMass || b.IsPointMass) throw new NotImplementedException();
				vx1 = 1.0/(max.Precision + a.Precision);
				mx1 = vx1*(max.MeanTimesPrecision + a.MeanTimesPrecision);
				vx2 = 1.0/(max.Precision + b.Precision);
				mx2 = vx2*(max.MeanTimesPrecision + b.MeanTimesPrecision);
			}
			logw1 = max.GetLogAverageOf(a);
			logPhi1 = MMath.NormalCdfLn((mx1 - mb) / Math.Sqrt(vx1 + vb));
			logu1 = Gaussian.GetLogProb(mx1, mb, vx1 + vb);

			logw2 = max.GetLogAverageOf(b);
			logPhi2 = MMath.NormalCdfLn((mx2 - ma) / Math.Sqrt(vx2 + va));
			logu2 = Gaussian.GetLogProb(mx2, ma, vx2 + va);

			logz = MMath.LogSumExp(logw1 + logPhi1, logw2 + logPhi2);
		}
#endif

        /// <summary>EP message to <c>val1</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val1</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val1</c> as the random arguments are varied. The formula is <c>proj[p(val1) sum_(max,val2) p(max,val2) factor(max,val1,val2)]/p(val1)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian AAverageConditional([SkipIfUniform] Gaussian max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            if (a.IsPointMass)
                return Gaussian.Uniform();
            if (max.IsUniform())
                return Gaussian.Uniform();
            if (!b.IsProper())
                throw new ImproperMessageException(b);
            double m1, v1, m2, v2;
            a.GetMeanAndVariance(out m1, out v1);
            b.GetMeanAndVariance(out m2, out v2);

            double logw1, alpha1, vx1, mx1;
            double logw2, alpha2, vx2, mx2;
            double logz;
            ComputeStats(max, a, b, out logz, out logw1, out alpha1, out vx1, out mx1,
                         out logw2, out alpha2, out vx2, out mx2);
            double w1 = Math.Exp(logw1 - logz);
            double w2 = Math.Exp(logw2 - logz);
            // the posterior is a mixture model with weights exp(logw1-logz), exp(logw2-logz) and distributions
            // N(a; mx1, vx1) phi((a - m2)/sqrt(v2)) / phi((mx1 - m2)/sqrt(vx1 + v2))
            // N(a; m1, v1) phi((mx2 - a)/sqrt(vx2)) / phi((mx2 - m1)/sqrt(vx2 + v1))
            // the moments of the posterior are computed via the moments of these two components.
            double mc1 = mx1;
            if(alpha1 != 0) // avoid 0*infinity
                mc1 += alpha1 * vx1;
            alpha2 = -alpha2;
            double mc2 = m1;
            if(alpha2 != 0) // avoid 0*infinity
                mc2 += alpha2 * v1;
            // z2 = (mx2 - m1) / Math.Sqrt(vx2 + v1)
            // logw2 = MMath.NormalCdfLn(z2);
            // alpha2 = -Math.Exp(Gaussian.GetLogProb(mx2, m1, vx2 + v1) - logw2);
            //        = -1/sqrt(vx2+v1)/NormalCdfRatio(z2)
            // m1 + alpha2*v1 = sqrt(vx2+v1)*(m1/sqrt(vx2+v1) - v1/(vx2+v1)/NormalCdfRatio(z2))
            //                = sqrt(vx2+v1)*(mx2/sqrt(vx2+v1) - z2 - v1/(vx2+v1)/NormalCdfRatio(z2))
            //                = sqrt(vx2+v1)*(mx2/sqrt(vx2+v1) - dY/Y)  if vx2=0
            double z2 = 0, Y = 0, dY = 0;
            if (vx2 == 0)
            {
                z2 = (mx2 - m1) / Math.Sqrt(vx2 + v1);
                if (z2 < -100)
                {
                    Y = MMath.NormalCdfRatio(z2);
                    dY = MMath.NormalCdfMomentRatio(1, z2);
                    mc2 = mx2 - Math.Sqrt(v1) * dY / Y;
                }
            }
            double m = w1 * mc1 + w2 * mc2;
            double beta1;
            if (alpha1 == 0)
                beta1 = 0;  // avoid 0*infinity
            else
            {
                double r1 = (mx1 - m2) / (vx1 + v2);
                beta1 = alpha1 * (alpha1 + r1);
            }
            double beta2;
            if (alpha2 == 0)
                beta2 = 0;  // avoid 0*infinity
            else
            {
                double r2 = (mx2 - m1) / (vx2 + v1);
                beta2 = alpha2 * (alpha2 - r2);
            }
            double vc1 = vx1 * (1 - vx1 * beta1);
            double vc2;
            if (vx2 == 0 && z2 < -100)
            {
                // beta2 = alpha2 * (alpha2 - z2/sqrt(v1))
                // vc2 = v1 - v1^2 * alpha2 * (alpha2 - z2/sqrt(v1))
                //     = v1 - v1*dY/Y^2
                //     =approx v1/z2^2
                // posterior E[x^2] = v - v*dY/Y^2 + v*dY^2/Y^2 = v - v*dY/Y^2*(1 - dY) = v + v*z*dY/Y = v*d2Y/Y
                //vc2 = v1 * (1 - dY / (Y * Y));
                double d2Y = 2 * MMath.NormalCdfMomentRatio(2, z2);
                double dYiY = dY / Y;
                vc2 = v1 * (d2Y / Y - dYiY * dYiY);
            }
            else if (beta2 == 0)
            {
                vc2 = v1;
            }
            else
            {
                vc2 = v1 * (1 - v1 * beta2);
            }
            double diff = mc1 - mc2;
            double v = w1 * vc1 + w2 * vc2 + w1 * w2 * diff * diff;
            Gaussian result = new Gaussian(m, v);
            result.SetToRatio(result, a, ForceProper);
            if (Double.IsNaN(result.Precision) || Double.IsNaN(result.MeanTimesPrecision))
                throw new InferRuntimeException($"result is NaN.  max={max}, a={a}, b={b}");
            return result;
        }

        /// <summary>EP message to <c>val1</c>.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val1</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val1</c> as the random arguments are varied. The formula is <c>proj[p(val1) sum_(val2) p(val2) factor(max,val1,val2)]/p(val1)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian AAverageConditional(double max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            return AAverageConditional(Gaussian.PointMass(max), a, b);
        }

        /// <summary>EP message to <c>val1</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>The outgoing EP message to the <c>val1</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val1</c> as the random arguments are varied. The formula is <c>proj[p(val1) sum_(max) p(max) factor(max,val1,val2)]/p(val1)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static Gaussian AAverageConditional([SkipIfUniform] Gaussian max, [Proper] Gaussian a, double b)
        {
            return AAverageConditional(max, a, Gaussian.PointMass(b));
        }

        /// <summary>EP message to <c>val1</c>.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Constant value for <c>val2</c>.</param>
        /// <returns>The outgoing EP message to the <c>val1</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>val1</c> conditioned on the given values.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        public static Gaussian AAverageConditional(double max, [Proper] Gaussian a, double b)
        {
            return AAverageConditional(Gaussian.PointMass(max), a, Gaussian.PointMass(b));
        }

        /// <summary>EP message to <c>val2</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val2</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val2</c> as the random arguments are varied. The formula is <c>proj[p(val2) sum_(max,val1) p(max,val1) factor(max,val1,val2)]/p(val2)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian BAverageConditional([SkipIfUniform] Gaussian max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            return AAverageConditional(max, b, a);
        }

        /// <summary>EP message to <c>val2</c>.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Incoming message from <c>val1</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val2</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val2</c> as the random arguments are varied. The formula is <c>proj[p(val2) sum_(val1) p(val1) factor(max,val1,val2)]/p(val2)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="a" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian BAverageConditional(double max, [Proper] Gaussian a, [Proper] Gaussian b)
        {
            return AAverageConditional(max, b, a);
        }

        /// <summary>EP message to <c>val2</c>.</summary>
        /// <param name="max">Incoming message from <c>max</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val2</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>val2</c> as the random arguments are varied. The formula is <c>proj[p(val2) sum_(max) p(max) factor(max,val1,val2)]/p(val2)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="max" /> is not a proper distribution.</exception>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian BAverageConditional([SkipIfUniform] Gaussian max, double a, [Proper] Gaussian b)
        {
            return AAverageConditional(max, b, a);
        }

        /// <summary>EP message to <c>val2</c>.</summary>
        /// <param name="max">Constant value for <c>max</c>.</param>
        /// <param name="a">Constant value for <c>val1</c>.</param>
        /// <param name="b">Incoming message from <c>val2</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <returns>The outgoing EP message to the <c>val2</c> argument.</returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>val2</c> conditioned on the given values.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="b" /> is not a proper distribution.</exception>
        public static Gaussian BAverageConditional(double max, double a, [Proper] Gaussian b)
        {
            return AAverageConditional(max, b, a);
        }
    }

    [FactorMethod(new string[] { "max", "a", "b" }, typeof(Math), "Max", typeof(double), typeof(double))]
    [Quality(QualityBand.Preview)]
    public static class MaxGammaOp
    {
        public static TruncatedGamma MaxAverageConditional(double a, Gamma b)
        {
            return new TruncatedGamma(b, a, double.PositiveInfinity);
        }

        public static TruncatedGamma MaxAverageConditional(Gamma a, double b)
        {
            return new TruncatedGamma(a, b, double.PositiveInfinity);
        }

        public static Gamma AAverageConditional(TruncatedGamma max, double b)
        {
            if (!max.IsPointMass)
                throw new ArgumentException("max is not a point mass");
            return Gamma.PointMass(max.Point);
        }

        public static Gamma BAverageConditional(TruncatedGamma max, double a)
        {
            if (!max.IsPointMass)
                throw new ArgumentException("max is not a point mass");
            return Gamma.PointMass(max.Point);
        }
    }
}
