/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;
    using MicrosoftResearch.Infer.Utils;

    public static class PointEstimator
    {
        /// <summary>
        /// Copy a value and make the downward message a point mass during inference.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="value"></param>
        /// <returns><paramref name="value"/></returns>
        public static T Forward<T>([IsReturned] T value)
        {
            return value;
        }

        [ParameterNames("forward", "value", "scale")]
        public static T Forward2<T>([IsReturned] T value, double scale)
        {
            return value;
        }
    }

    [FactorMethod(typeof(PointEstimator), "Forward2<>", typeof(double), typeof(double), typeof(double), Default = true)]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForward2Op<T>
    {
        [Skip]
        public static double LogEvidenceRatio()
        {
            return 0.0;
        }

        [SkipIfAllUniform]
        public static TDist ForwardAverageConditional<TDist>(TDist value, [Fresh] TDist to_value, TDist result)
            where TDist : SettableToProduct<TDist>, CanGetMean<double>, HasPoint<double>
        {
            result.SetToProduct(value, to_value);
            double mean = result.GetMean();
            result.Point = mean;
            return result;
        }

        public static Gamma ValueAverageConditional(double scale, Gamma forward, [Cancels] Gamma to_forward)
        {
            if (forward.IsPointMass)
                return forward;
            double x = to_forward.Point;
            double dlogp, ddlogp;
            forward.GetDerivatives(x, out dlogp, out ddlogp);
            ddlogp *= scale;
            if (double.IsNaN(dlogp) || double.IsNaN(ddlogp))
                throw new Exception("nan");
            return Gamma.FromDerivatives(x, dlogp, ddlogp, true);
        }

        public static Beta ValueAverageConditional(double scale, Beta forward, [Cancels] Beta to_forward)
        {
            if (forward.IsPointMass)
                return forward;
            double p = to_forward.Point;
            double dlogp, ddlogp;
            forward.GetDerivatives(p, out dlogp, out ddlogp);
            ddlogp *= scale;
            if (double.IsNaN(dlogp) || double.IsNaN(ddlogp))
                throw new Exception("nan");
            return Beta.FromDerivatives(p, dlogp, ddlogp, true);
        }

        public static Gaussian ValueAverageConditional(double scale, Gaussian forward, [Cancels] Gaussian to_forward)
        {
            if (forward.Precision < 0)
            {
                double deriv = forward.MeanTimesPrecision - forward.Precision * to_forward.Point;
                forward = Gaussian.FromDerivatives(to_forward.Point, deriv, 0, false);
            }
            // deriv = mp - prec*point
            // return fromNatural(deriv + scale*prec*point, scale*prec)
            return Gaussian.FromNatural(forward.MeanTimesPrecision + (scale - 1) * forward.Precision * to_forward.Point, scale * forward.Precision);
        }
    }

    //[FactorMethod(typeof(PointEstimator), "Forward<>", Default = false)]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForwardOp_Mode<T>
    {
        [Skip]
        public static double LogEvidenceRatio()
        {
            return 0.0;
        }

        public static double LogEvidenceRatio2<TDist>(TDist forward, TDist value, TDist to_forward)
            where TDist : CanGetLogAverageOf<TDist>
        {
            return value.GetLogAverageOf(forward) - to_forward.GetLogAverageOf(forward);
        }

        [SkipIfAllUniform]
        public static TDist ForwardAverageConditional<TDist>(TDist forward, TDist value, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMode<T>
        {
            result.SetToProduct(forward, value);
            result.Point = result.GetMode();
            return result;
        }

        public static TDist ValueAverageConditional<TDist>([IsReturned] TDist forward, TDist result)
            where TDist : SettableTo<TDist>
        {
            result.SetTo(forward);
            return result;
        }
    }

    [FactorMethod(typeof(PointEstimator), "Forward<>", Default = false)]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForwardOp_Mean<T>
    {
        [Skip]
        public static double LogEvidenceRatio()
        {
            return 0.0;
        }

        public static double LogEvidenceRatio2<TDist>(TDist forward, TDist value, TDist to_forward)
            where TDist : CanGetLogAverageOf<TDist>
        {
            return value.GetLogAverageOf(forward) - to_forward.GetLogAverageOf(forward);
        }

        [SkipIfAllUniform]
        public static TDist ForwardAverageConditional<TDist>([NoInit] TDist forward, TDist value, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMean<T>
        {
            result.SetToProduct(forward, value);
            result.Point = result.GetMean();
            return result;
        }

        public static TDist ValueAverageConditional<TDist>([IsReturned] TDist forward, TDist result)
            where TDist : SettableTo<TDist>
        {
            result.SetTo(forward);
            return result;
        }
    }

    /// <summary>
    /// Generic delegate with 2 out parameters
    /// </summary>
    /// <typeparam name="T1">Type of first argument</typeparam>
    /// <typeparam name="T2">Type of second argument</typeparam>
    /// <typeparam name="T3">Type of third argument</typeparam>
    /// <param name="arg1">First argument</param>
    /// <param name="arg2">Second argument</param>
    /// <param name="arg3">Third argument</param>
    public delegate void ActionOut2<in T1, T2, T3>(T1 arg1, out T2 arg2, out T3 arg3);

    /// <summary>
    /// Generic delegate with 2 out parameters
    /// </summary>
    /// <typeparam name="T1">Type of first argument</typeparam>
    /// <typeparam name="T2">Type of second argument</typeparam>
    /// <typeparam name="T3">Type of third argument</typeparam>
    /// <typeparam name="T4">Type of fourth argument</typeparam>
    /// <param name="arg1">First argument</param>
    /// <param name="arg2">Second argument</param>
    /// <param name="arg3">Third argument</param>
    /// <param name="arg4">Fourth argument</param>
    public delegate void ActionOut2<in T1, in T2, T3, T4>(T1 arg1, T2 arg2, out T3 arg3, out T4 arg4);

    public class SecantBufferData<T>
        where T : HasPoint<double>, SettableToProduct<T>, new()
    {
        /// <summary>
        /// -log10(Tolerance) is the number of digits of precision required when estimating the Hessian
        /// </summary>
        public static double Tolerance = 1e-2;
        public static bool debug;
        public static List<double[]> history = new List<double[]>();
        public static int maxPoints = 100;
        public static bool evictLowestWeight;
        /// <summary>
        /// Index of the most recently added point.
        /// </summary>
        public int position;
        public List<double> prevPoints = new List<double>();
        public List<double> prevDerivs = new List<double>();
        public List<double> prevDeriv2s = new List<double>();
        public List<double> uncertainties = new List<double>();
        public T dist;
        public bool hasDist, hasDeriv2;
        public double nextPoint;
        public double nextNegativeHessian = 1e10;
        public double scaleUp = 2;
        public double stepsize;
        Func<T, bool> isProper;
        ActionOut2<T, double, double, double> getDerivatives;
        Func<double, double, double, bool, T> fromDerivatives;
        /// <summary>
        /// The mode in the original parameter space.  (Not neccesarily the same space as getDerivatives.)
        /// </summary>
        Func<T, double> getMode;
        /// <summary>
        /// The mode in the transformed parameter space.
        /// </summary>
        Func<T, double> getModeTransformed;

        public SecantBufferData(
            Func<T,bool> isProper,
            Func<T,double> getMode, 
            Func<T,double> getModeTransformed,
            ActionOut2<T, double, double, double> getDerivatives, 
            Func<double, double, double, bool, T> fromDerivatives)
        {
            this.isProper = isProper;
            this.getMode = getMode;
            this.getModeTransformed = getModeTransformed;
            this.getDerivatives = getDerivatives;
            this.fromDerivatives = fromDerivatives;
        }

        public void Add(double point, double deriv, double deriv2)
        {
            double uncertainty = 1;
            if (prevPoints.Count > 0)
            {
                double diff = point - prevPoints[position];
                uncertainty = uncertainties[position] / 2 + diff * diff;
            }
            if (prevPoints.Count < maxPoints)
            {
                position = prevPoints.Count;
                prevPoints.Add(0.0);
                prevDerivs.Add(0.0);
                prevDeriv2s.Add(0.0);
                uncertainties.Add(0.0);
            }
            else if(evictLowestWeight)
            {
                // find the point with smallest weight
                int excluded = position;
                position = 0;
                double maxDist = double.NegativeInfinity;
                for (int i = 0; i < prevPoints.Count; i++)
                {
                    if (i == excluded)
                        continue;
                    double diff = point - prevPoints[i];
                    double dist = diff * diff + uncertainties[i];
                    if (dist > maxDist)
                    {
                        maxDist = dist;
                        position = i;
                    }
                }
            }
            else
            {
                position++;
                if (position == prevPoints.Count)
                    position = 0;
            }
            prevPoints[position] = point;
            prevDerivs[position] = deriv;
            prevDeriv2s[position] = deriv2;
            uncertainties[position] = uncertainty;
            if(debug) 
                history.Add(new double[] { point, deriv, deriv2, uncertainty, position });
        }

        T Product(T a, T b)
        {
            T result = new T();
            result.SetToProduct(a, b);
            return result;
        }

        public bool GetDeriv2(double currPoint, double currDeriv, double currDeriv2, double prevDeriv2, 
            out double newDeriv, out double newDeriv2)
        {
            int count = prevPoints.Count;
            if (count == 0)
            {
                newDeriv = currDeriv;
                newDeriv2 = 0;
                return false;
            }
            else
            {
                bool rescale = (currDeriv2 < 0) && prevDeriv2s.All(x => (x < 0));
                var derivs = Util.ArrayInit(count, i =>
                {
                    if (rescale)
                        return prevDerivs[i] * (currDeriv2 / prevDeriv2s[i]);
                    else
                        return prevDerivs[i];
                });
                var points = prevPoints;
                var weights = Util.ArrayInit(count, i => 1.0/(1e-8 + 1000*Math.Pow(currPoint - prevPoints[i],2) + uncertainties[i]));
                //var weights = Util.ArrayInit(count, i => Math.Exp(-50* (Math.Pow(currPoint - prevPoints[i], 2) + uncertainties[i])));
                //weights = Util.ArrayInit(count, i => 1.0);
                double slope, offset;
                bool ok = LinearRegression(points, derivs, weights, prevDeriv2, out slope, out offset);
                if (ok)
                {
                    newDeriv = offset + slope * currPoint;
                    newDeriv2 = slope;
                    return true;
                }
                else
                {
                    //Debug.WriteLine("delta = {0:g4} minDelta = {1:g4}", Math.Abs(currPoint - prevPoints[0]), minDelta);
                    newDeriv = currDeriv;
                    newDeriv2 = 0;
                    return false;
                }
            }
        }

        public void SetNextPoint(T forward, T value)
        {
            T marginal = Product(forward, value);
            if (!isProper(marginal))
                marginal = value;
            this.nextPoint = getMode(marginal);
        }

        public void SetNextPoint(double currPoint, T forward, T value)
        {
            SecantBufferData<T> buffer = this;
            T currDist = forward;
            if (currDist.IsPointMass)
            {
                buffer.nextPoint = currDist.Point;
                return;
            }
            if (value.IsPointMass)
            {
                buffer.nextPoint = value.Point;
                return;
            }
            // cannot use buffer.nextPoint as currPoint since Forward could be initialized by user
            double currDeriv, currDeriv2;
            getDerivatives(currDist, currPoint, out currDeriv, out currDeriv2);
            double prevSign = 0;
            double prevDeriv = 0;
            double prevPoint = 0;
            double prevStep = 0;
            if (buffer.prevPoints.Count > 0)
            {
                prevPoint = buffer.prevPoints[buffer.position];
                prevDeriv = buffer.prevDerivs[buffer.position];
                prevStep = currPoint - prevPoint;
                prevSign = Math.Sign(prevStep);
            }
            buffer.Add(currPoint, currDeriv, currDeriv2);
            bool useSecant = (buffer.prevPoints.Count > 1);
            useSecant = false; // TEMPORARY
            if (useSecant)
            {
                // suppose prevDeriv = currDist.MeanTimesPrecision - currDist.Precision * prevPoint
                // then (currDeriv - prevDeriv)/(currPoint - prevPoint) is accurate to tolerance iff
                // currDist.Precision * abs(currPoint - prevPoint) >= abs(currDist.MeanTimesPrecision) * 1e-16/tolerance
                // or equivalently abs(currPoint - prevPoint) >= abs(currDist.GetMean()) * 1e-16/tolerance
                // provided currDist.Precision > 0
                double prevDeriv2;
                double newDeriv;
                double newDeriv2;
                if (buffer.hasDeriv2)
                {
                    getDerivatives(buffer.dist, currPoint, out newDeriv, out newDeriv2);
                    prevDeriv2 = newDeriv2;
                }
                else
                {
                    prevDeriv2 = currDeriv2;
                }
                useSecant = buffer.GetDeriv2(currPoint, currDeriv, currDeriv2, prevDeriv2, out newDeriv, out newDeriv2);
                if (useSecant)
                {
                    bool rejectLargeDeriv2 = false;
                    if (rejectLargeDeriv2 && newDeriv2 >= currDeriv2 + 1e-10)
                    {
                        if (debug)
                            Debug.WriteLine("rejecting newDeriv2={0} when currDeriv2={1}", newDeriv2, currDeriv2);
                        // in this case, the secant step would try to accelerate the normal step, which is dangerous.
                        // use the normal step instead.
                        buffer.hasDist = false;
                        useSecant = false;
                        //buffer.nextPoint = (currDist * value).GetMean();
                        //buffer.dist = currDist;
                        //buffer.hasDist = true;
                    }
                    else
                    {
                        // secant step
                        T newDist = fromDerivatives(currPoint, newDeriv, newDeriv2, false);
                        if (!isProper(newDist))
                        {
                            if (debug)
                                Debug.WriteLine("improper distribution: {0}", newDist);
                            //buffer.hasDist = false;
                            useSecant = false;
                        }
                        else
                        {
                            if(debug)
                                Debug.WriteLine("provided gradient = {0:g4} estimated gradient = {1:g4} provided hessian = {2:g4} estimated hessian = {3:g4}", currDeriv, newDeriv, currDeriv2, newDeriv2);
                            buffer.nextPoint = getMode(Product(newDist, value));
                            buffer.dist = newDist;
                            buffer.hasDist = true;
                            buffer.hasDeriv2 = true;
                            if (double.IsNaN(buffer.nextPoint))
                                throw new Exception("nextPoint is NaN");
                            //if (!isProper(currDist))
                            //    currDeriv2 = newDeriv2;
                        }
                    }
                }
            }
            if (!useSecant)
            {
                if (buffer.prevPoints.Count <= 1)
                    buffer.hasDist = false;
                T marginal = new T();
                if (buffer.hasDeriv2)
                {
                    double newDeriv, newDeriv2;
                    getDerivatives(buffer.dist, currPoint, out newDeriv, out newDeriv2);
                    T newDist = fromDerivatives(currPoint, currDeriv, newDeriv2, true);
                    buffer.dist = newDist;
                    buffer.hasDist = true;
                    marginal.SetToProduct(newDist, value);
                    if(debug)
                        Debug.WriteLine("provided gradient = {0:g4} provided hessian = {1:g4} using previous hessian = {2:g4}", currDeriv, currDeriv2, newDeriv2);
                }
                else
                {
                    bool useForward = false;
                    // TEMPORARY
                    buffer.hasDist = true;
                    if (!buffer.hasDist)
                    {
                        marginal.SetToProduct(forward, value);
                        if (isProper(marginal))
                        {
                            useForward = true;
                            buffer.hasDist = true;
                            double newDeriv, newDeriv2;
                            getDerivatives(marginal, currPoint, out newDeriv, out newDeriv2);
                            buffer.nextNegativeHessian = -newDeriv2;
                            if (debug)
                                Debug.WriteLine("provided gradient = {0:g4} provided hessian = {1:g4} using forward = {2}", currDeriv, currDeriv2, forward);
                        }
                    }
                    if (!useForward)
                    {
                        T newDist2 = fromDerivatives(currPoint, currDeriv, -buffer.nextNegativeHessian, true);
                        marginal.SetToProduct(newDist2, value);
                        double newDeriv, newDeriv2;
                        getDerivatives(marginal, currPoint, out newDeriv, out newDeriv2);
                        if (buffer.prevPoints.Count > 1)
                        {
                            double currStep = getModeTransformed(marginal) - currPoint;
                            double currSign = Math.Sign(currStep);
                            // this assumes getMode is actually the mode
                            //double currSign = Math.Sign(newDeriv);
                            if (debug)
                                Debug.WriteLine("prevSign = {0}, currSign = {1}", prevSign, currSign);
                            if (prevSign != 0 && currSign != 0)
                            {
                                if (prevSign != currSign)
                                {
                                    //double slope;
                                    //bool ok = TrySecant(currDeriv, prevDeriv, currPoint, prevPoint, out slope);
                                    //if (ok)
                                    //{
                                    //    // secant rule
                                    //    buffer.nextNegativeHessian = -slope;
                                    //}
                                    //else
                                    //{
                                    //    // Rprop heuristic
                                    //    buffer.nextNegativeHessian *= 2;
                                    //}
                                    buffer.scaleUp = 1.2;
                                    buffer.stepsize /= 2;
                                }
                                else
                                {
                                    // prevSign == currSign
                                    //if (Math.Abs(currStep) > Math.Abs(prevStep))
                                    //{
                                    //    // Since the hessian has not been changed yet, the only way this can happen is if
                                    //    // the gradient magnitude has increased.  In that case, scale the hessian to compensate.
                                    //    buffer.nextNegativeHessian *= Math.Abs(currStep) / Math.Abs(prevStep);
                                    //    if (debug)
                                    //        Debug.WriteLine("step increased from {0} to {1}", Math.Abs(prevStep), Math.Abs(currStep));
                                    //}
                                    //buffer.nextNegativeHessian /= buffer.scaleUp;
                                    buffer.stepsize *= buffer.scaleUp;
                                }
                            }
                        }
                        else
                        {
                            //buffer.nextNegativeHessian /= buffer.scaleUp;
                            if (currPoint == 0)
                                buffer.stepsize = Math.Abs(currDeriv);
                            else
                                buffer.stepsize = Math.Abs(currPoint) * 1e-4;
                        }
                        // Rprop
                        if(newDeriv != 0)
                            buffer.nextNegativeHessian = Math.Abs(newDeriv) / buffer.stepsize;
                        T newDist = fromDerivatives(currPoint, currDeriv, -buffer.nextNegativeHessian, true);
                        if(debug)
                            Debug.WriteLine("inventing {0}, nextDeriv2 = {1}, currPoint = {2}, currDeriv = {3}", newDist, -buffer.nextNegativeHessian, currPoint, currDeriv);
                        buffer.dist = newDist;
                        buffer.hasDist = true;
                        marginal.SetToProduct(newDist, value);
                    }
                }
                buffer.nextPoint = getMode(marginal);
            }
        }

        private bool TrySecant(double currDeriv, double prevDeriv, double currPoint, double prevPoint, out double slope)
        {
            double numerError = MachineError(currDeriv) + MachineError(prevDeriv);
            double denomError = MachineError(currPoint) + MachineError(prevPoint);
            double numer = currDeriv - prevDeriv;
            double denom = currPoint - prevPoint;
            slope = numer / denom;
            double slopeUpper = (numer + numerError) / (denom - denomError);
            double slopeLower = (numer - numerError) / (denom + denomError);
            // upper and lower might swap due to a negative slope, so take absolute difference
            bool ok = (Math.Abs(slopeUpper - slopeLower) < MachineError(slope) * 100);
            return ok;
        }

        private double MachineError(double x)
        {
            return Math.Abs(x) * 1e-16;
        }

        public static int[] GetSortedIndices(double[] keys)
        {
            int[] indices = Util.ArrayInit(keys.Length, i => i);
            Array.Sort(keys, indices);
            return indices;
        }

        public static bool IsSingular(PositiveDefiniteMatrix gram, double minDelta)
        {
            LowerTriangularMatrix gramChol = new LowerTriangularMatrix(gram.Rows, gram.Cols);
            gramChol.SetToCholesky(gram);
            for (int j = 0; j < gram.Rows; j++)
            {
                if (gramChol[j, j] < minDelta)
                {
                    return true;
                }
            }
            return false;
        }

        public static bool LinearRegression(IList<double> x, IList<double> y, IList<double> weight, double slopeEstimate, out double slope, out double offset)
        {
            // indices[0] is the index with largest weight
            int[] indices = GetSortedIndices(weight.Select(w => -w).ToArray());
            double maxAbsY = 0;
            double maxAbsX = 0;
            double machEps = 1e-16;
            for (int i = 0; i < indices.Length; i++)
            {
                int index = indices[i];
                if(debug)
                    Debug.WriteLine("x[{0}] = {1}, y = {2}", index, x[index], y[index]);
                maxAbsX = Math.Max(maxAbsX, Math.Abs(x[index]));
                maxAbsY = Math.Max(maxAbsY, Math.Abs(y[index]));
                if (i < 1)
                    continue;
                // construct matrix of [1 x]
                DenseVector yVector = DenseVector.Zero(i + 1);
                Matrix X1 = new Matrix(i+1, 2);
                for (int j = 0; j <= i; j++)
                {
                    X1[j, 0] = 1;
                    X1[j, 1] = x[indices[j]];
                    yVector[j] = y[indices[j]];
                }
                // Because of double-precision storage, y is only accurate to within abs(y)*1e-16.
                // If every y has a standard deviation of e, and the gramian has singular value s, 
                // then the standard deviation of slope is e/s.
                // We want this to be less than abs(slope)*Tolerance, therefore we need
                // e/s <= abs(slope)*Tolerance
                // s >= e/abs(slope)/Tolerance
                double minDelta = maxAbsY * machEps / (Math.Abs(slopeEstimate) * Tolerance);
                // x is only accurate to within abs(x)*machEps.
                //minDelta = Math.Max(minDelta, maxAbsX * machEps / Tolerance);
                double minDeltaSqr = minDelta * minDelta;
                //Debug.WriteLine("minDeltaSqr = {0}", minDeltaSqr);
                DenseVector coeffs = DenseVector.Zero(2);
                double minSingularValueSqr = coeffs.SetToLeastSquares(yVector, X1);
                if (minSingularValueSqr > minDeltaSqr)
                {
                    offset = coeffs[0];
                    slope = coeffs[1];
                    if (double.IsNaN(slope))
                        throw new Exception("slope is nan");
                    if(debug)
                        Debug.WriteLine("count = {0}, slope = {1}, minSingularValueSqr = {2}, minDeltaSqr = {3}", i + 1, slope, minSingularValueSqr, minDeltaSqr);
                    return true;
                }
            }
            slope = double.NaN;
            offset = double.NaN;
            return false;
        }

        public static void WriteHistory(string path)
        {
            WriteHistory(history, path);
        }

        internal static void WriteHistory(List<double[]> history, string path)
        {
            Matrix m = new Matrix(history.Count, (history.Count == 0) ? 0 : history[0].Length);
            int rowIndex = 0;
            foreach (var row in history)
            {
                for (int colIndex = 0; colIndex < row.Length; colIndex++)
                {
                    m[rowIndex, colIndex] = row[colIndex];
                }
                rowIndex++;
            }
            using (MatlabWriter writer = new MatlabWriter(path))
            {
                writer.Write("history", m);
            }
        }
    }

    [FactorMethod(typeof(PointEstimator), "Forward<>", typeof(double), Default = true)]
    [Buffers("buffer")]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForwardOp_Secant
    {
        [Skip]
        public static SecantBufferData<Gaussian> BufferInit()
        {
            return new SecantBufferData<Gaussian>(g => g.IsProper(), g => g.GetMode(), g => g.GetMode(),
                Gaussian.GetDerivatives, Gaussian.FromDerivatives);
        }

        // must compute new forward here since Forward method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static SecantBufferData<Gaussian> Buffer([NoInit] Gaussian forward, Gaussian value, [Cancels] Gaussian to_forward, SecantBufferData<Gaussian> buffer)
        {
            // cannot use buffer.nextPoint as currPoint since Forward could be initialized by user
            if (to_forward.IsPointMass)
            {
                double currPoint = to_forward.Point;
                buffer.SetNextPoint(currPoint, forward, value);
            }
            else
            {
                //throw new Exception("to_forward is not a point mass.  PointEstimator variables should be initialized to a point mass.");
                buffer.SetNextPoint(forward, value);
            }
            return buffer;
        }

        public static Gaussian ForwardAverageConditional([IgnoreDependency] Gaussian forward, [IgnoreDependency] Gaussian value, [SkipIfUniform] SecantBufferData<Gaussian> buffer, Gaussian result)
        {
            result.Point = buffer.nextPoint;
            return result;
        }

        public static Gaussian ValueAverageConditional([SkipIfUniform] Gaussian forward, [NoInit] SecantBufferData<Gaussian> buffer)
        {
            if (buffer.hasDist)
            {
                return buffer.dist;
            }
            else
            {
                return forward;
            }
        }
    }

    [FactorMethod(typeof(PointEstimator), "Forward<>", typeof(double), Default = true)]
    [Buffers("buffer")]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForwardOp_SecantGamma
    {
        public static bool UseMean;

        public static void GammaGetDerivatives(Gamma dist, double x, out double mdlogp, out double xxddlogp)
        {
            // the density is measured at 1/x, the first derivative is negated to keep the same meaning of second derivative
            double a = dist.Shape - 1;
            mdlogp = dist.Rate - a * x;
            xxddlogp = -a;
        }

        public static Gamma GammaFromDerivatives(double x, double mdlogp, double xxddlogp, bool forceProper)
        {
            // the density is measured at 1/x
            double ddlogp = xxddlogp * x * x;
            return Gamma.FromDerivatives(1 / x, -mdlogp, ddlogp, forceProper);
        }

        [Skip]
        public static SecantBufferData<Gamma> BufferInit()
        {
            Func<Gamma, double> getMode = g => g.GetMode();
            Func<Gamma, double> getMean = g => g.GetMean();
            Func<Gamma, double> getModeFunc = UseMean ? getMean : getMode;
            return new SecantBufferData<Gamma>(g => g.IsProper(), getModeFunc, g => 1/getModeFunc(g),
                GammaGetDerivatives, GammaFromDerivatives);
        }

        // must compute new forward here since Forward method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform("forward","value")]
        public static SecantBufferData<Gamma> Buffer([NoInit] Gamma forward, Gamma value, [Cancels] Gamma to_forward, SecantBufferData<Gamma> buffer)
        {
            if (to_forward.IsPointMass)
            {
                // cannot use buffer.nextPoint as currPoint since Forward could be initialized by user
                double currPoint = 1 / to_forward.Point;
                buffer.SetNextPoint(currPoint, forward, value);
            }
            else
            {
                //throw new Exception("to_forward is not a point mass.  PointEstimator variables should be initialized to a point mass.");
                buffer.SetNextPoint(forward, value);
            }
            return buffer;
        }

        public static Gamma ForwardAverageConditional([IgnoreDependency] Gamma forward, [IgnoreDependency] Gamma value, [RequiredArgument] SecantBufferData<Gamma> buffer, Gamma result)
        {
            result.Point = buffer.nextPoint;
            return result;
        }

        public static Gamma ValueAverageConditional([SkipIfUniform] Gamma forward, [RequiredArgument] SecantBufferData<Gamma> buffer)
        {
            if (buffer.hasDist)
            {
                return buffer.dist;
            }
            else
            {
                return forward;
            }
        }
    }
}
