/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using MicrosoftResearch.Infer.Distributions;
using MicrosoftResearch.Infer.Maths;
using MicrosoftResearch.Infer.Utils;

namespace MicrosoftResearch.Infer.Factors
{
    public class VectorSecantBufferData
    {
        /// <summary>
        /// -log10(Tolerance) is the number of digits of precision required when estimating the Hessian
        /// </summary>
        public static double Tolerance = 1e-8;
        public static bool debug;
        public static List<double[]> history = new List<double[]>();
        public static int maxPoints = 100;
        public static bool evictLowestWeight;
        /// <summary>
        /// Index of the most recently added point.
        /// </summary>
        public int position;
        public List<DenseVector> prevPoints = new List<DenseVector>();
        public List<DenseVector> prevDerivs = new List<DenseVector>();
        public List<PositiveDefiniteMatrix> prevNegativeHessians = new List<PositiveDefiniteMatrix>();
        public List<double> uncertainties = new List<double>();
        public List<int> recency = new List<int>();
        public int timestep;
        public VectorGaussian dist;
        public bool hasDist, hasDeriv2;
        public DenseVector nextPoint;
        public PositiveDefiniteMatrix nextNegativeHessian;

        public void Add(DenseVector point, DenseVector deriv, PositiveDefiniteMatrix negativeHessian)
        {
            double uncertainty = 1;
            if (prevPoints.Count > 0)
            {
                var diff = point - prevPoints[position];
                uncertainty = uncertainties[position] / 2 + diff.Inner(diff);
            }
            if (prevPoints.Count < maxPoints)
            {
                position = prevPoints.Count;
                // the values used here don't matter since they will be overwritten below
                prevPoints.Add(DenseVector.Zero(point.Count));
                prevDerivs.Add(DenseVector.Zero(point.Count));
                prevNegativeHessians.Add(new PositiveDefiniteMatrix(point.Count, point.Count));
                uncertainties.Add(0.0);
                recency.Add(0);
            }
            else if (evictLowestWeight)
            {
                // find the point with smallest weight
                int excluded = position;
                position = 0;
                double maxDist = double.NegativeInfinity;
                for (int i = 0; i < prevPoints.Count; i++)
                {
                    if (i == excluded)
                        continue;
                    var diff = point - prevPoints[i];
                    double dist = diff.Inner(diff) + uncertainties[i];
                    if (dist > maxDist)
                    {
                        maxDist = dist;
                        position = i;
                    }
                }
            }
            else
            {
                // find a matching point
                double minDist = point.Inner(point) * 1e-16;
                bool foundMatch = false;
                for (int i = 0; i < prevPoints.Count; i++)
                {
                    var diff = point - prevPoints[i];
                    double dist = diff.Inner(diff);
                    if (dist < minDist)
                    {
                        position = i;
                        foundMatch = true;
                        break;
                    }
                }
                if (!foundMatch)
                {
                    position++;
                    if (position == prevPoints.Count)
                        position = 0;
                }
            }
            prevPoints[position].SetTo(point);
            prevDerivs[position].SetTo(deriv);
            prevNegativeHessians[position].SetTo(negativeHessian);
            uncertainties[position] = uncertainty;
            recency[position] = timestep;
            timestep++;
            if (debug)
                history.Add(point.Concat(deriv).Concat(new double[] { uncertainty, position }).ToArray());
        }

        public bool GetDeriv2(DenseVector currPoint, DenseVector currDeriv, PositiveDefiniteMatrix currNegativeHessian, PositiveDefiniteMatrix prevNegativeHessian, out DenseVector newDeriv, out PositiveDefiniteMatrix negativeHessian)
        {
            int dimension = currPoint.Count;
            negativeHessian = new PositiveDefiniteMatrix(dimension, dimension);
            if (prevPoints.Count == 0)
            {
                newDeriv = currDeriv;
                return false;
            }
            var weights = Util.ArrayInit(prevPoints.Count, i =>
            {
                var diff = currPoint - prevPoints[i];
                return 1.0 / (1e-8 + diff.Inner(diff) + uncertainties[i]);
            });
            weights = Util.ArrayInit(prevPoints.Count, i => (double)recency[i]);
            // indices[0] is the index with largest weight
            int[] indices = SecantBufferData<Gaussian>.GetSortedIndices(weights.Select(w => -w).ToArray());
            double maxAbsY = 0;
            double maxAbsX = 0;
            double machEps = 1e-16;
            Matrix derivs = null;
            List<Vector> scaledDerivs = new List<Vector>();
            LowerTriangularMatrix L = new LowerTriangularMatrix(dimension, dimension);
            for (int i = 0; i < prevDerivs.Count; i++)
            {
                Vector scaledDeriv = RescaleDeriv(prevDerivs[i], prevNegativeHessians[i], currNegativeHessian, L);
                scaledDerivs.Add(scaledDeriv);
            }
            L.SetToCholesky(prevNegativeHessian);
            double minPrevEigenvalue = double.PositiveInfinity;
            double prevLogDet = 0;
            for (int i = 0; i < dimension; i++)
            {
                minPrevEigenvalue = Math.Min(minPrevEigenvalue, L[i, i]*L[i,i]);
                prevLogDet += 2*Math.Log(L[i, i]);
            }
            double prevEigenvalue = Math.Exp(prevLogDet / dimension);
            for (int count = 1; count <= prevPoints.Count; count++)
            {
                int index2 = indices[count-1];
                if(debug)
                Debug.WriteLine("x[{0}] = {1}, y = {2}", index2, prevPoints[index2], scaledDerivs[index2]);
                derivs = new Matrix(count, dimension);
                var X1 = new Matrix(count, dimension + 1);
                for (int i = 0; i < count; i++)
                {
                    int index = indices[i];
                    for (int j = 0; j < dimension; j++)
                    {
                        X1[i, j] = prevPoints[index][j];
                        derivs[i, j] = scaledDerivs[index][j];
                        if (i == count - 1)
                        {
                            maxAbsX = Math.Max(maxAbsX, Math.Abs(prevPoints[index][j]));
                            maxAbsY = Math.Max(maxAbsY, Math.Abs(scaledDerivs[index][j]));
                        }
                    }
                    X1[i, dimension] = 1;
                }
                if (count < dimension + 1)
                    continue;
                double minDelta = maxAbsY * machEps / (prevEigenvalue * Tolerance);
                //double minDelta = maxAbsX * machEps / Tolerance;
                // needed for SetToLyapunovSymmetric
                //minDelta = Math.Max(minDelta, 1e-8);
                double minDeltaSqr = minDelta * minDelta;
                bool useLeastSquares = false;
                if (useLeastSquares)
                {
                    Matrix coeffs = new Matrix(dimension+1, dimension);
                    double minSingularValueSqr = coeffs.SetToLeastSquares(derivs, X1);
                    if (minSingularValueSqr > minDeltaSqr)
                    {
                        var offset = Vector.Zero(dimension);
                        offset.SetToSubarray(coeffs.GetRow(dimension), 0);
                        negativeHessian.SetToSubmatrix(coeffs, 0, 0);
                        negativeHessian.Symmetrize();
                        newDeriv = DenseVector.Zero(dimension);
                        newDeriv.SetToProduct(currPoint, negativeHessian);
                        newDeriv.SetToSum(newDeriv, offset);
                        negativeHessian.Scale(-1);
                        if (negativeHessian.Any(double.IsNaN))
                            throw new Exception("negativeHessian is nan");
                        if (negativeHessian.IsPositiveDefinite())
                            return true;
                        negativeHessian.SetAllElementsTo(0.0);
                        // fall through
                    }
                }
                else
                {
                    Matrix V = new Matrix(dimension + 1, dimension + 1);
                    Matrix US = (Matrix)X1.Clone();
                    V.SetToRightSingularVectors(US);
                    Matrix VtP = new Matrix(dimension + 1, dimension);
                    VtP.SetToSubmatrix(V.Transpose(), 0, 0);
                    bool singular = false;
                    Vector singularValuesSqr = Vector.Zero(US.Cols);
                    for (int k = 0; k < US.Cols; k++)
                    {
                        double singularValueSqr = 0;
                        for (int j = 0; j < US.Rows; j++)
                        {
                            singularValueSqr += US[j, k] * US[j, k];
                        }
                        singularValuesSqr[k] = singularValueSqr;
                        if (singularValueSqr <= minDeltaSqr)
                            singular = true;
                    }
                    if (!singular)
                    {
                        if(debug)
                            Debug.WriteLine("singularValuesSqr = {0}, minDeltaSqr = {1}", singularValuesSqr, minDeltaSqr);
                        try
                        {
                            // vec(S*V'*Y'*P*U + U'*P'*Y*V*S) - (kron(U'*D*U,S*S) + kron(S*S,U'*D*U))*vec(U'*C*U)
                            Matrix VPYUS = VtP * derivs.Transpose() * US;
                            VPYUS.Symmetrize();
                            VPYUS.Scale(2);
                            var S2 = new PositiveDefiniteMatrix(US.Cols, US.Cols);
                            S2.SetDiagonal(singularValuesSqr);
                            var VtD = new Matrix(V.Cols, V.Cols);
                            VtD.SetSubmatrix(0, 0, VtP);
                            var VtDV = new PositiveDefiniteMatrix(V.Cols, V.Cols);
                            VtDV.SetToOuter(VtD);
                            Matrix UtCU = new Maths.Matrix(dimension + 1, dimension + 1);
                            SetToLyapunovSymmetric(UtCU, S2, VtDV, VPYUS);
                            Matrix C = V * UtCU * V.Transpose();
                            C.Symmetrize();
                            if (true)
                            {
                                // iterate again on the residual for higher accuracy
                                var residual = new Matrix(count, dimension);
                                residual.SetToSubmatrix(X1 * C, 0, 0);
                                residual.SetToDifference(derivs, residual);
                                VPYUS.SetToProduct(VtP * residual.Transpose(), US);
                                VPYUS.Symmetrize();
                                VPYUS.Scale(2);
                                SetToLyapunovSymmetric(UtCU, S2, VtDV, VPYUS);
                                Matrix C2 = V * UtCU * V.Transpose();
                                C2.Symmetrize();
                                C.SetToSum(C, C2);
                            }

                            var offset = Vector.Zero(dimension);
                            offset.SetToSubarray(C.GetColumn(dimension), 0);
                            negativeHessian.SetToSubmatrix(C, 0, 0);
                            newDeriv = DenseVector.Zero(dimension);
                            newDeriv.SetToProduct(currPoint, negativeHessian);
                            newDeriv.SetToSum(newDeriv, offset);
                            negativeHessian.Scale(-1);
                            if (negativeHessian.Any(double.IsNaN))
                                throw new Exception("negativeHessian is nan");
                            if (negativeHessian.IsPositiveDefinite())
                            {
                                var diff = newDeriv - currDeriv;
                                var derivError = diff.Inner(diff);
                                var derivMagnitude = currDeriv.Inner(currDeriv);
                                if (debug)
                                    Debug.WriteLine("derivError/derivMagnitude = {0}", derivError / derivMagnitude);
                                if (derivError < derivMagnitude * 1e-10)
                                    return true;
                                //Debug.WriteLine(StringUtil.JoinColumns("negativeHessian would be ", negativeHessian));
                            }
                            else if (debug)
                            {
                                Debug.WriteLine("estimated negativeHessian is not positive definite");
                            }
                            negativeHessian.SetAllElementsTo(0.0);
                            // do not expand the neighborhood
                            break;
                        }
                        catch (PositiveDefiniteMatrixException) { }
                    }
                }
            }
            newDeriv = DenseVector.Zero(dimension);
            if (this.hasDeriv2)
                return false;
            // fit each dimension independently
            for (int j = 0; j < dimension; j++)
            {
                double slope, offset;
                bool ok = SecantBufferData<Gaussian>.LinearRegression(
                    prevPoints.Select(x => x[j]).ToArray(), 
                    scaledDerivs.Select(x => x[j]).ToArray(),
                    weights, -prevNegativeHessian[j,j], out slope, out offset);
                if (ok && slope <= 0)
                {
                    newDeriv[j] = offset + slope * currPoint[j];
                    negativeHessian[j, j] = -slope;
                }
                else
                {
                    newDeriv = currDeriv;
                    return false;
                }
            }
            return true;
        }

        public void SetNextPoint(VectorGaussian forward, VectorGaussian value)
        {
            var marginal = forward * value;
            if (!marginal.IsProper())
                marginal = value;
            this.nextPoint = marginal.GetMean();
        }

        public void SetNextPoint(DenseVector currPoint, VectorGaussian forward, VectorGaussian value)
        {
            VectorSecantBufferData buffer = this;
            VectorGaussian currDist = forward;
            if (currDist.IsPointMass)
            {
                buffer.nextPoint = (DenseVector)currDist.Point;
                return;
            }
            // deriv of -0.5*prec*x^2+pm*x
            // is -prec*x + pm
            DenseVector currDeriv = (DenseVector)(currDist.MeanTimesPrecision - currDist.Precision * currPoint);
            PositiveDefiniteMatrix currNegativeHessian = currDist.Precision;
            Vector prevSign = null;
            if (buffer.prevPoints.Count > 0)
            {
                var prevPoint = buffer.prevPoints[buffer.position];
                var step = currPoint - prevPoint;
                prevSign = Normalize(step);
            }
            buffer.Add(currPoint, currDeriv, currNegativeHessian);
            bool useSecant = (buffer.prevPoints.Count > 1);
            useSecant = false; // TEMPORARY
            if (useSecant)
            {
                // suppose prevDeriv = currDist.MeanTimesPrecision - currDist.Precision * prevPoint
                // then (currDeriv - prevDeriv)/(currPoint - prevPoint) is accurate to tolerance iff
                // currDist.Precision * abs(currPoint - prevPoint) >= abs(currDist.MeanTimesPrecision) * 1e-16/tolerance
                // or equivalently abs(currPoint - prevPoint) >= abs(currDist.GetMean()) * 1e-16/tolerance
                // provided currDist.Precision > 0
                DenseVector newDeriv;
                PositiveDefiniteMatrix negativeHessian;
                PositiveDefiniteMatrix prevNegativeHessian;
                if (buffer.hasDeriv2)
                    prevNegativeHessian = buffer.dist.Precision;
                else
                    prevNegativeHessian = currNegativeHessian;
                useSecant = buffer.GetDeriv2(currPoint, currDeriv, currNegativeHessian, prevNegativeHessian, out newDeriv, out negativeHessian);
                if (useSecant)
                {
                    if ((currDist.Precision - negativeHessian).IsPositiveDefinite())
                    {
                        buffer.hasDist = false;
                        useSecant = false;
                    }
                    else
                    {
                        // secant step
                        //buffer.nextPoint = (buffer.prevPoint * currDeriv - currPoint * buffer.prevDeriv) / (currDeriv - buffer.prevDeriv);
                        VectorGaussian newDist = VectorGaussian.FromDerivatives(currPoint, newDeriv, negativeHessian, false);
                        if (!newDist.IsProper())
                        {
                            useSecant = false;
                        }
                        else
                        {
                            if (debug)
                                Debug.WriteLine("provided gradient = {0:g4} estimated gradient = {1:g4} provided hessian = {2:g4} estimated hessian = {3:g4}", currDeriv, newDeriv, currNegativeHessian, negativeHessian);
                            buffer.nextPoint = (newDist * value).GetMean();
                            buffer.dist = newDist;
                            buffer.hasDist = true;
                            buffer.hasDeriv2 = true;
                            if (buffer.nextPoint.Any(double.IsNaN))
                                throw new Exception("nextPoint is NaN");
                        }
                    }
                }
            }
            if (!useSecant)
            {
                if (buffer.prevPoints.Count <= 1)
                    buffer.hasDist = false;
                VectorGaussian marginal = new VectorGaussian(currDeriv.Count);
                if (buffer.hasDeriv2)
                {
                    VectorGaussian newDist = VectorGaussian.FromDerivatives(currPoint, currDeriv, buffer.dist.Precision, true);
                    buffer.dist = newDist;
                    buffer.hasDist = true;
                    marginal.SetToProduct(newDist, value);
                    if (debug)
                        Debug.WriteLine("provided gradient = {0:g4} provided hessian = {1:g4} using previous hessian = {2:g4}", currDeriv, currNegativeHessian, buffer.dist.Precision);
                }
                else
                {
                    bool useForward = false;
                    // TEMPORARY
                    buffer.hasDist = true;
                    if (!buffer.hasDist)
                    {
                        marginal.SetToProduct(forward, value);
                        if (marginal.IsProper())
                        {
                            useForward = true;
                            if (debug)
                                Debug.WriteLine("provided gradient = {0:g4} provided hessian = {1:g4} using forward = {2}", currDeriv, currNegativeHessian, forward);
                        }
                    }
                    if (!useForward)
                    {
                        if (buffer.nextNegativeHessian == null)
                            buffer.nextNegativeHessian = PositiveDefiniteMatrix.IdentityScaledBy(currDeriv.Count, 1e6);
                        //buffer.nextNegativeHessian.SetToIdentityScaledBy(50);
                        VectorGaussian newDist = VectorGaussian.FromDerivatives(currPoint, currDeriv, buffer.nextNegativeHessian, true);
                        if (debug)
                            Debug.WriteLine("inventing {0}", newDist);
                        // the Hessian should shrink to allow movement in new directions
                        buffer.nextNegativeHessian.Scale(0.9);
                        buffer.dist = newDist;
                        buffer.hasDist = true;
                        marginal.SetToProduct(newDist, value);
                        if (buffer.prevPoints.Count > 1)
                        {
                            var step = marginal.GetMean() - currPoint;
                            var stepIsZero = step.EqualsAll(0);
                            if (!stepIsZero)
                            {
                                var currSign = Normalize(step);
                                var inner = prevSign.Inner(currSign);
                                double scaleUp = 1 + (1 - inner) * 2;
                                double scaleDown = 1 + (inner + 1) / 2;
                                if (debug)
                                {
                                    //double quad = buffer.nextNegativeHessian.QuadraticForm(currSign, currSign);
                                    //Debug.WriteLine("inner = {0}, scaleUp = {1}, scaleDown = {2}, quad = {3}, prevSign = {4}, currSign = {5}", inner, scaleUp, scaleDown, quad, prevSign, currSign);
                                    //Debug.WriteLine(StringUtil.JoinColumns("before update = ", buffer.nextNegativeHessian));         
                                }
                                // sum and diff are perpendicular
                                var sum = Normalize(prevSign + currSign);
                                var diff = Normalize(currSign - prevSign);
                                ScaleAlongDirection(buffer.nextNegativeHessian, sum, 1 / scaleDown);
                                ScaleAlongDirection(buffer.nextNegativeHessian, diff, scaleUp);
                                if (debug)
                                {
                                    //Debug.WriteLine("quad after update = {0}", buffer.nextNegativeHessian.QuadraticForm(currSign, currSign));
                                    //Debug.WriteLine(StringUtil.JoinColumns("after update = ", buffer.nextNegativeHessian));
                                }
                            }
                        }
                    }
                }
                buffer.nextPoint = marginal.GetMean();
            }
        }

        /// <summary>
        /// Modifies A so that result.QuadraticForm(b) = scale*A.QuadraticForm(b) while remaining positive definite.
        /// </summary>
        /// <param name="A"></param>
        /// <param name="b">Must be unit length.</param>
        /// <param name="scale">Must be non-negative.</param>
        private void ScaleAlongDirection(PositiveDefiniteMatrix A, Vector b, double scale)
        {
            Debug.Assert(scale >= 0);
            double quad = A.QuadraticForm(b, b);
            var Ab = A * b;
            A.SetToSumWithOuter(A, -1, Ab, b);
            A.SetToSumWithOuter(A, -1, b, Ab);
            A.SetToSumWithOuter(A, (scale + 1)*quad, b, b);
            A.Symmetrize();
            if (!A.IsPositiveDefinite())
                throw new Exception();
            bool check = false;
            if (check)
            {
                double newQuad = A.QuadraticForm(b, b);
                Debug.Assert(MMath.AbsDiff(newQuad, scale * quad, 1e-8) < 1e-8);
            }
        }

        private Vector Normalize(Vector vector)
        {
            double inner = vector.Inner(vector);
            if(inner > 0) 
                vector.Scale(1 / Math.Sqrt(inner));
            return vector;
        }

        private Vector RescaleDeriv(DenseVector prevDeriv, PositiveDefiniteMatrix prevNegativeHessian, PositiveDefiniteMatrix currNegativeHessian, LowerTriangularMatrix L)
        {
            if (L.SetToCholesky(prevNegativeHessian))
            {
                //scale = (currNegativeHessian / prevNegativeHessians[0]);
                var scaledDeriv = prevDeriv.Clone();
                scaledDeriv.PredivideBy(L);
                scaledDeriv.PredivideByTranspose(L);
                return currNegativeHessian * scaledDeriv;
            }
            else
            {
                return prevDeriv;
            }
        }

        /// <summary>
        /// Set X to satisfy (A*X*B + B*X*A = C).
        /// </summary>
        /// <param name="X">Will be symmetric</param>
        /// <param name="A">Must be symmetric</param>
        /// <param name="B">Must be symmetric</param>
        /// <param name="C">Must be symmetric</param>
        public static void SetToLyapunovSymmetric(Matrix X, PositiveDefiniteMatrix A, PositiveDefiniteMatrix B, Matrix C)
        {
            Assert.IsTrue(C.SymmetryError() == 0);
            Assert.IsTrue(B.SymmetryError() == 0);
            Assert.IsTrue(A.SymmetryError() == 0);
            Vector vecC = Vector.FromList(C);
            int dimension = A.Rows;
            PositiveDefiniteMatrix AkronB = new PositiveDefiniteMatrix(dimension * dimension, dimension * dimension);
            AkronB.SetTo(A.Kronecker(B) + B.Kronecker(A));
            if (true)
            {
                // since B is not positive definite, add a correction to avoid a singular matrix
                // kron(I-B,I-B)
                var IminusB = PositiveDefiniteMatrix.Identity(dimension) - B;
                AkronB.SetToSum(AkronB, IminusB.Kronecker(IminusB));
            }
            vecC.PredivideBy(AkronB);
            X.SetTo(vecC.ToArray());
            X.Symmetrize();
            bool check = false;
            if (check)
            {
                var C2 = A * X * B + B * X * A;
                Assert.IsTrue(C2.MaxDiff(C) < 1e-8);
            }
        }

        public static void WriteHistory(string path)
        {
            SecantBufferData<Gaussian>.WriteHistory(history, path);
        }
    }

    [FactorMethod(typeof(PointEstimator), "Forward<>", typeof(Vector), Default = true)]
    [Buffers("buffer")]
    [Quality(QualityBand.Experimental)]
    public static class PointEstimatorForwardOp_VectorSecant
    {
        [Skip]
        public static VectorSecantBufferData BufferInit()
        {
            return new VectorSecantBufferData();
        }

        // must compute new forward here since Forward method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static VectorSecantBufferData Buffer([NoInit] VectorGaussian forward, [Cancels] VectorGaussian value, [Cancels] VectorGaussian to_forward, VectorSecantBufferData buffer)
        {
            // cannot use buffer.nextPoint as currPoint since Forward could be initialized by user
            if (to_forward.IsPointMass)
            {
                var currPoint = (DenseVector)to_forward.Point;
                buffer.SetNextPoint(currPoint, forward, value);
            }
            else
            {
                //throw new Exception("to_forward is not a point mass.  PointEstimator variables should be initialized to a point mass.");
                buffer.SetNextPoint(forward, value);
            }
            return buffer;
        }

        public static VectorGaussian ForwardAverageConditional([IgnoreDependency] VectorGaussian forward, [IgnoreDependency] VectorGaussian value, [SkipIfUniform] VectorSecantBufferData buffer, VectorGaussian result)
        {
            result.Point = buffer.nextPoint;
            return result;
        }

        public static VectorGaussian ValueAverageConditional([SkipIfUniform] VectorGaussian forward, VectorSecantBufferData buffer)
        {
            if (buffer.hasDist)
            {
                return buffer.dist;
            }
            else
            {
                return forward;
            }
        }

        public static Gaussian ValueAverageConditional2([IsReturned] Gaussian forward)
        {
            return forward;
        }
    }
}
