/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

using System;
using MicrosoftResearch.Infer.Distributions;
using MicrosoftResearch.Infer.Maths;

namespace MicrosoftResearch.Infer.Factors
{
    /// <summary>Provides outgoing messages for the following factors:<list type="bullet"><item><description><see cref="Factor.Product(double, double)" /></description></item><item><description><see cref="Factor.Ratio(double, double)" /></description></item></list>, given random arguments to the function.</summary>
    [FactorMethod(typeof(Factor), "Product", typeof(double), typeof(double))]
    [FactorMethod(new string[] { "A", "Product", "B" }, typeof(Factor), "Ratio", typeof(double), typeof(double))]
    [Quality(QualityBand.Experimental)]
    public static class WrappedGaussianProductOp
    {
        /// <summary>EP message to <c>a</c>.</summary>
        /// <param name="Product">Incoming message from <c>product</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>a</c> as the random arguments are varied. The formula is <c>proj[p(a) sum_(product) p(product) factor(product,a,b)]/p(a)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Product" /> is not a proper distribution.</exception>
        public static WrappedGaussian AAverageConditional([SkipIfUniform] WrappedGaussian Product, double B, WrappedGaussian result)
        {
            result.Period = Product.Period / B;
            result.Gaussian = GaussianProductOp.AAverageConditional(Product.Gaussian, B);
            result.Normalize();
            return result;
        }

        /// <summary>EP message to <c>b</c>.</summary>
        /// <param name="Product">Incoming message from <c>product</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>b</c> as the random arguments are varied. The formula is <c>proj[p(b) sum_(product) p(product) factor(product,a,b)]/p(b)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Product" /> is not a proper distribution.</exception>
        public static WrappedGaussian BAverageConditional([SkipIfUniform] WrappedGaussian Product, double A, WrappedGaussian result)
        {
            return AAverageConditional(Product, A, result);
        }

        /// <summary>EP message to <c>a</c>.</summary>
        /// <param name="Product">Constant value for <c>product</c>.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>a</c> conditioned on the given values.</para>
        /// </remarks>
        public static WrappedGaussian AAverageConditional(double Product, double B, WrappedGaussian result)
        {
            if (B == 0)
            {
                if (Product != 0)
                    throw new AllZeroException();
                result.SetToUniform();
            }
            else
                result.Point = Product / B;
            result.Normalize();
            return result;
        }

        // ----------------------------------------------------------------------------------------------------------------------
        // VMP
        // ----------------------------------------------------------------------------------------------------------------------

        /// <summary>Evidence message for VMP.</summary>
        /// <param name="product">Incoming message from <c>product</c>.</param>
        /// <returns>Zero.</returns>
        /// <remarks>
        ///   <para>In Variational Message Passing, the evidence contribution of a deterministic factor is zero. Adding up these values across all factors and variables gives the log-evidence estimate for VMP.</para>
        /// </remarks>
        [Skip]
        public static double AverageLogFactor(WrappedGaussian product)
        {
            return 0.0;
        }

        /// <summary>VMP message to <c>product</c>.</summary>
        /// <param name="A">Incoming message from <c>a</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>product</c> as the random arguments are varied. The formula is <c>proj[sum_(a) p(a) factor(product,a,b)]</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="A" /> is not a proper distribution.</exception>
        public static WrappedGaussian ProductAverageLogarithm([SkipIfUniform] WrappedGaussian A, double B, WrappedGaussian result)
        {
            double m, v;
            A.Gaussian.GetMeanAndVariance(out m, out v);
            result.Gaussian.SetMeanAndVariance(B * m, B * B * v);
            double period = B * A.Period;
            if (period != result.Period)
            {
                double ratio = period / result.Period;
                double intRatio = Math.Round(ratio);
                if (Math.Abs(ratio - intRatio) > result.Period * 1e-4)
                    throw new ArgumentException("B*A.Period (" + period + ") is not a multiple of result.Period (" + result.Period + ")");
                // if period is a multiple of result.Period, then wrapping to result.Period is equivalent to first wrapping to period, then to result.Period.
            }
            result.Normalize();
            return result;
        }

        /// <summary>VMP message to <c>product</c>.</summary>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="B">Incoming message from <c>b</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>product</c> as the random arguments are varied. The formula is <c>proj[sum_(b) p(b) factor(product,a,b)]</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="B" /> is not a proper distribution.</exception>
        public static WrappedGaussian ProductAverageLogarithm(double A, [SkipIfUniform] WrappedGaussian B, WrappedGaussian result)
        {
            return ProductAverageLogarithm(B, A, result);
        }

        /// <summary>VMP message to <c>a</c>.</summary>
        /// <param name="Product">Incoming message from <c>product</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>a</c> with <c>product</c> integrated out. The formula is <c>sum_product p(product) factor(product,a,b)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Product" /> is not a proper distribution.</exception>
        public static WrappedGaussian AAverageLogarithm([SkipIfUniform] WrappedGaussian Product, double B, WrappedGaussian result)
        {
            if (Product.IsPointMass)
                return AAverageLogarithm(Product.Point, B, result);
            return AAverageConditional(Product, B, result);
        }

        /// <summary>VMP message to <c>b</c>.</summary>
        /// <param name="Product">Incoming message from <c>product</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="A">Constant value for <c>a</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>b</c> with <c>product</c> integrated out. The formula is <c>sum_product p(product) factor(product,a,b)</c>.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Product" /> is not a proper distribution.</exception>
        public static WrappedGaussian BAverageLogarithm([SkipIfUniform] WrappedGaussian Product, double A, WrappedGaussian result)
        {
            return AAverageLogarithm(Product, A, result);
        }

        /// <summary>VMP message to <c>a</c>.</summary>
        /// <param name="Product">Constant value for <c>product</c>.</param>
        /// <param name="B">Constant value for <c>b</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the factor viewed as a function of <c>a</c> conditioned on the given values.</para>
        /// </remarks>
        public static WrappedGaussian AAverageLogarithm(double Product, double B, WrappedGaussian result)
        {
            return AAverageConditional(Product, B, result);
        }
    }
}
