/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

#define SpecializeArrays
#define MinimalGenericTypeParameters

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using System.Collections.Generic;
    using MicrosoftResearch.Infer;
    using MicrosoftResearch.Infer.Collections;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

    /// <summary>Provides outgoing messages for the following factors:<list type="bullet"><item><description><see cref="Factor.Replicate{T}(T, int)" /></description></item><item><description><see cref="Factor.ReplicateWithMarginal{T}(T, int, out T)" /></description></item></list>, given random arguments to the function.</summary>
    /// <typeparam name="T">The type of the variable being replicated.</typeparam>
    [FactorMethod(typeof(Factor), "Replicate<>", Default = true)]
    [FactorMethod(typeof(Factor), "ReplicateWithMarginal<>", Default = true)]
    [Quality(QualityBand.Mature)]
    public static class ReplicateGibbsOp<T>
    {
        /// <summary>Evidence message for EP.</summary>
        /// <param name="Uses">Incoming message from <c>Uses</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <returns>Logarithm of the factor's contribution the EP model evidence.</returns>
        /// <remarks>
        ///   <para>The formula for the result is <c>log(sum_(Uses,Def) p(Uses,Def) factor(Uses,Def,Count) / sum_Uses p(Uses) messageTo(Uses))</c>. Adding up these values across all factors and variables gives the log-evidence estimate for EP.</para>
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Uses" /> is not a proper distribution.</exception>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static double LogEvidenceRatio<TDist>([SkipIfAllUniform] IList<TDist> Uses, T Def)
            where TDist : IDistribution<T>
        {
            return 0.0;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static double GibbsEvidence<TDist>(IList<TDist> Uses, T Def)
            where TDist : IDistribution<T>
        {
            return 0.0;
        }

#if false
		public static T UsesGibbs<TDist>([SkipIfUniform] GibbsMarginal<TDist, T> marginal, int resultIndex, T result)
			where TDist : IDistribution<T>, Sampleable<T>
		{
			return marginal.LastSample;
		}
#elif false
		public static T UsesGibbs<TDist>([SkipIfUniform] GibbsMarginal<TDist, T> marginal, TDist def, int resultIndex, T result)
			where TDist : IDistribution<T>, Sampleable<T>
		{
			return marginal.LastSample;
		}
		public static T UsesGibbs<TDist>([SkipIfUniform] GibbsMarginal<TDist, T> marginal, T def, int resultIndex, T result)
			where TDist : IDistribution<T>, Sampleable<T>
		{
			if (def is bool[]) {
				if (!Util.ValueEquals((bool[])(object)def, (bool[])(object)marginal.LastSample)) throw new Exception("gotcha");
			}
			else if (def is double[]) {
				if (!Util.ValueEquals((double[])(object)def, (double[])(object)marginal.LastSample)) throw new Exception("gotcha");
			} else if (!def.Equals(marginal.LastSample)) throw new Exception("gotcha");
			return marginal.LastSample;
		}
#else
        /// <summary />
        /// <param name="to_marginal" />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static T UsesGibbs<TDist>([SkipIfUniform] GibbsMarginal<TDist, T> to_marginal, TDist def, int resultIndex, T result)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            // This method must depend on Def, even though Def isn't used, in order to get the right triggers
            return to_marginal.LastSample;
        }

        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        public static T UsesGibbs([IsReturned] T def, int resultIndex, T result)
        {
            return def;
        }
#endif

#if true
        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        // until .NET 4
        public static TDist UsesGibbs<TDist>([IsReturned] TDist def, int resultIndex, TDist result)
            where TDist : IDistribution<T>
        {
            return def;
        }
#endif

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Uses" /> is not a proper distribution.</exception>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [MultiplyAll]
        public static TDist DefGibbs<TDist>(
            [SkipIfAllUniform] IList<TDist> Uses,
            TDist result)
            where TDist : IDistribution<T>, Sampleable<T>, SettableTo<TDist>, SettableToProduct<TDist>, SettableToRatio<TDist>
        {
            return ReplicateOp_NoDivide.DefAverageConditional(Uses, result);
        }

#if SpecializeArrays
        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Uses" /> is not a proper distribution.</exception>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [MultiplyAll]
        public static TDist DefGibbs<TDist>(
            [SkipIfAllUniform] TDist[] Uses,
            TDist result)
            where TDist : IDistribution<T>, Sampleable<T>, SettableToProduct<TDist>, SettableTo<TDist>
        {
            return ReplicateOp_NoDivide.DefAverageConditional(Uses, result);
        }
#endif

        /// <summary />
        /// <param name="to_marginal" />
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static T DefGibbs<TDist>([SkipIfUniform] GibbsMarginal<TDist, T> to_marginal, T result)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            return to_marginal.LastSample;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="Def">Incoming message from <c>Def</c>. Must be a proper distribution. If uniform, the result will be uniform.</param>
        /// <param name="to_marginal" />
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Def" /> is not a proper distribution.</exception>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Stochastic]
        [SkipIfAllUniform]
        public static GibbsMarginal<TDist, T> MarginalGibbs<TDist>(
            IList<TDist> Uses,
            [SkipIfUniform] TDist Def,
            GibbsMarginal<TDist, T> to_marginal)
            where TDist : IDistribution<T>, SettableToProduct<TDist>, SettableToRatio<TDist>, SettableTo<TDist>, Sampleable<T>
        {
            GibbsMarginal<TDist, T> result = to_marginal;
            TDist marginal = result.LastConditional;
            marginal.SetTo(Def);
            marginal = Distribution.SetToProductWithAll(marginal, Uses);
            result.LastConditional = marginal;
            // Allow a sample to be drawn from the last conditional, and add it to the sample
            // list and conditional list
            result.PostUpdate();
            return result;
        }

        /// <summary />
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <param name="to_marginal" />
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Stochastic] // must be labelled Stochastic to get correct schedule, even though it isn't Stochastic
        public static GibbsMarginal<TDist, T> MarginalGibbs<TDist>(
            T Def,
            GibbsMarginal<TDist, T> to_marginal)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            GibbsMarginal<TDist, T> result = to_marginal;
            TDist marginal = result.LastConditional;
            marginal.Point = Def;
            result.LastConditional = marginal;
            // Allow a sample to be drawn from the last conditional, and add it to the sample
            // list and conditional list
            result.PostUpdate();
            return result;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="to_marginal" />
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Stochastic] // must be labelled Stochastic to get correct schedule, even though it isn't Stochastic
        public static GibbsMarginal<TDist, T> MarginalGibbs<TDist>(
            T[] Uses,
            GibbsMarginal<TDist, T> to_marginal)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            GibbsMarginal<TDist, T> result = to_marginal;
            TDist marginal = result.LastConditional;
            if (Uses.Length != 1)
                throw new ArgumentException("Uses.Length (" + Uses.Length + ") != 1");
            marginal.Point = Uses[0];
            result.LastConditional = marginal;
            // Allow a sample to be drawn from the last conditional, and add it to the sample
            // list and conditional list
            result.PostUpdate();
            return result;
        }

        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static GibbsMarginal<TDist, T> MarginalGibbsInit<TDist>([IgnoreDependency] TDist def)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            return new GibbsMarginal<TDist, T>(def, 100, 1, true, true, true);
        }
    }

    /// <summary>Provides outgoing messages for <see cref="Factor.ReplicateWithMarginalGibbs{T}(T, int, int, int, out T, out T, out T)" />, given random arguments to the function.</summary>
    /// <typeparam name="T">The type of the replicated variable.</typeparam>
    [FactorMethod(typeof(Factor), "ReplicateWithMarginalGibbs<>")]
    [Buffers("sample", "conditional", "marginalEstimator", "sampleAcc", "conditionalAcc")]
    [Quality(QualityBand.Mature)]
    public static class ReplicateGibbsOp2<T>
    {
        /// <summary>Initialize the buffer <c>conditional</c>.</summary>
        /// <param name="to_marginal">Previous outgoing message to <c>marginal</c>.</param>
        /// <returns>Initial value of buffer <c>conditional</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static TDist ConditionalInit<TDist>([IgnoreDependency] TDist to_marginal)
            where TDist : ICloneable
        {
            return (TDist)to_marginal.Clone();
        }

        /// <summary>Update the buffer <c>conditional</c>.</summary>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static TDist Conditional<TDist>(T Def, TDist result)
            where TDist : HasPoint<T>
        {
            result.Point = Def;
            return result;
        }

        /// <summary>Update the buffer <c>conditional</c>.</summary>
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static TDist Conditional<TDist>(IList<TDist> Uses, [SkipIfAnyUniform] TDist Def, TDist result)
            where TDist : SettableTo<TDist>, SettableToProduct<TDist>
        {
            result.SetTo(Def);
            result = Distribution.SetToProductWithAll(result, Uses);
            return result;
        }

        /// <summary>Update the buffer <c>sample</c>.</summary>
        /// <param name="to_marginal">Previous outgoing message to <c>marginal</c>.</param>
        /// <param name="conditional">Buffer <c>conditional</c>.</param>
        /// <returns>New value of buffer <c>sample</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Stochastic]
        public static T Sample<TDist>([IgnoreDependency] TDist to_marginal, [Proper] TDist conditional)
            where TDist : IDistribution<T>, Sampleable<T>
        {
            return conditional.Sample();
        }

        /// <summary>Initialize the buffer <c>marginalEstimator</c>.</summary>
        /// <param name="to_marginal">Previous outgoing message to <c>marginal</c>.</param>
        /// <param name="burnIn">Constant value for <c>burnIn</c>.</param>
        /// <returns>Initial value of buffer <c>marginalEstimator</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static BurnInAccumulator<TDist> MarginalEstimatorInit<TDist>([IgnoreDependency] TDist to_marginal, int burnIn)
            where TDist : IDistribution<T>
        {
            Accumulator<TDist> est = (Accumulator<TDist>)ArrayEstimator.CreateEstimator<TDist, T>(to_marginal, true);
            return new BurnInAccumulator<TDist>(burnIn, 1, est);
        }

        /// <summary>Update the buffer <c>marginalEstimator</c>.</summary>
        /// <param name="conditional">Buffer <c>conditional</c>.</param>
        /// <param name="marginalEstimator">Buffer <c>marginalEstimator</c>.</param>
        /// <returns>New value of buffer <c>marginalEstimator</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        /// <typeparam name="TAcc">The type of the marginal estimator.</typeparam>
        public static TAcc MarginalEstimator<TDist, TAcc>([Proper] TDist conditional, TAcc marginalEstimator)
            where TAcc : Accumulator<TDist>
        {
            marginalEstimator.Add(conditional);
            return marginalEstimator;
        }

        /// <summary />
        /// <param name="marginalEstimator">Buffer <c>marginalEstimator</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static TDist MarginalGibbs<TDist>(BurnInAccumulator<TDist> marginalEstimator, TDist result)
        {
            return ((Estimator<TDist>)marginalEstimator.Accumulator).GetDistribution(result);
        }

        /// <summary>Initialize the buffer <c>sampleAcc</c>.</summary>
        /// <param name="to_samples">Previous outgoing message to <c>samples</c>.</param>
        /// <param name="burnIn">Constant value for <c>burnIn</c>.</param>
        /// <param name="thin">Constant value for <c>thin</c>.</param>
        /// <returns>Initial value of buffer <c>sampleAcc</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        public static Accumulator<T> SampleAccInit(ICollection<T> to_samples, int burnIn, int thin)
        {
            return new BurnInAccumulator<T>(burnIn, thin, new AccumulateIntoCollection<T>(to_samples));
        }

        /// <summary>Update the buffer <c>sampleAcc</c>.</summary>
        /// <param name="sample">Buffer <c>sample</c>.</param>
        /// <param name="sampleAcc">Buffer <c>sampleAcc</c>.</param>
        /// <returns>New value of buffer <c>sampleAcc</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        public static Accumulator<T> SampleAcc(T sample, Accumulator<T> sampleAcc)
        {
            sampleAcc.Add(sample);
            return sampleAcc;
        }

        /// <summary />
        /// <param name="sampleAcc">Buffer <c>sampleAcc</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TList">The type of the outgoing message.</typeparam>
        public static TList SamplesGibbs<TList>(Accumulator<T> sampleAcc, TList result)
            where TList : ICollection<T>
        {
            // do nothing since result was already modified by sampleAcc
            return result;
        }

        /// <summary>Initialize the buffer <c>conditionalAcc</c>.</summary>
        /// <param name="to_conditionals">Previous outgoing message to <c>conditionals</c>.</param>
        /// <param name="burnIn">Constant value for <c>burnIn</c>.</param>
        /// <param name="thin">Constant value for <c>thin</c>.</param>
        /// <returns>Initial value of buffer <c>conditionalAcc</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static Accumulator<TDist> ConditionalAccInit<TDist>(ICollection<TDist> to_conditionals, int burnIn, int thin)
        {
            return new BurnInAccumulator<TDist>(burnIn, thin, new AccumulateIntoCollection<TDist>(to_conditionals));
        }

        /// <summary>Update the buffer <c>conditionalAcc</c>.</summary>
        /// <param name="conditional">Buffer <c>conditional</c>.</param>
        /// <param name="conditionalAcc">Buffer <c>conditionalAcc</c>.</param>
        /// <returns>New value of buffer <c>conditionalAcc</c>.</returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static Accumulator<TDist> ConditionalAcc<TDist>(TDist conditional, Accumulator<TDist> conditionalAcc)
            where TDist : ICloneable
        {
            conditionalAcc.Add((TDist)conditional.Clone());
            return conditionalAcc;
        }

        /// <summary />
        /// <param name="conditionalAcc">Buffer <c>conditionalAcc</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        /// <typeparam name="TDistList">The type of the outgoing message.</typeparam>
        public static TDistList ConditionalsGibbs<TDist, TDistList>(Accumulator<TDist> conditionalAcc, TDistList result)
            where TDistList : ICollection<TDist>
        {
            // do nothing since result was already modified by Acc
            return result;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static double GibbsEvidence<TDist>(IList<TDist> Uses, T Def)
            where TDist : IDistribution<T>
        {
            return 0.0;
        }

        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        public static T UsesGibbs([IsReturned] T def, int resultIndex, T result)
        {
            return def;
        }

        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="sample">Buffer <c>sample</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static T UsesGibbs<TDist>(TDist def, T sample, int resultIndex, T result)
            where TDist : IDistribution<T>
        {
            // This method must depend on Def, even though Def isn't used, in order to get the right triggers
            return sample;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>.</param>
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static TDist UsesGibbs<TDist>(
            [IgnoreDependency] ICollection<TDist> Uses,
            [IsReturned] TDist Def,
            int resultIndex, TDist result)
            where TDist : IDistribution<T>, Sampleable<T>, SettableTo<TDist>, SettableToProduct<TDist>, SettableToRatio<TDist>
        {
            if (resultIndex < 0 || resultIndex >= Uses.Count)
                throw new ArgumentOutOfRangeException("resultIndex");
            if (Uses.Count > 1)
                throw new ArgumentException("Uses.Count > 1");
            result.SetTo(Def);
            return result;
        }

        /// <summary />
        /// <param name="Def">Incoming message from <c>Def</c>.</param>
        /// <param name="resultIndex">Index of the <c>Uses</c> for which a message is desired.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        [Skip]
        public static TDist UsesGibbsInit<TDist>([IgnoreDependency] TDist Def, int resultIndex)
            where TDist : ICloneable
        {
            return (TDist)Def.Clone();
        }

        /// <summary />
        /// <param name="def">Incoming message from <c>Def</c>.</param>
        /// <param name="sample">Buffer <c>sample</c>.</param>
        /// <returns />
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        public static T DefGibbs<TDist>(TDist def, [IsReturned] T sample)
            where TDist : IDistribution<T>
        {
            return sample;
        }

        /// <summary />
        /// <param name="Uses">Incoming message from <c>Uses</c>. Must be a proper distribution. If all elements are uniform, the result will be uniform.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para />
        /// </remarks>
        /// <exception cref="ImproperMessageException">
        ///   <paramref name="Uses" /> is not a proper distribution.</exception>
        /// <typeparam name="TDist">The type of the distribution over the replicated variable.</typeparam>
        //[MultiplyAll]
        public static TDist DefGibbs<TDist>(
            [SkipIfAllUniform] IList<TDist> Uses,
            TDist result)
            where TDist : IDistribution<T>, Sampleable<T>, SettableTo<TDist>, SettableToProduct<TDist>, SettableToRatio<TDist>
        {
            result.SetToUniform();
            result = Distribution.SetToProductWithAll(result, Uses);
            return result;
        }
    }

}
