/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace MicrosoftResearch.Infer.Factors
{
    using System;
    using System.Diagnostics;
    using MicrosoftResearch.Infer.Distributions;
    using MicrosoftResearch.Infer.Maths;

    [Quality(QualityBand.Preview)]
    public class VariablePointOpBase
    {
        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        public static double LogEvidenceRatio<TDist>(TDist use, TDist def, TDist to_marginal)
            where TDist : CanGetLogAverageOf<TDist>
        {
            //return def.GetLogAverageOf(to_marginal);
            return def.GetLogAverageOf(use) - use.GetLogAverageOf(to_marginal);
        }

        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        [Skip]
        public static TDist MarginalAverageConditionalInit<TDist>([IgnoreDependency] TDist def)
            where TDist : ICloneable
        {
            return (TDist)def.Clone();
        }

        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        public static TDist UseAverageConditional<TDist>([IsReturned] TDist to_marginal)
        {
            return to_marginal;
        }

        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        public static TDist DefAverageConditional<TDist>([IsReturned] TDist to_marginal)
        {
            return to_marginal;
        }

        [Skip]
        public static double AverageLogFactor()
        {
            return 0.0;
        }

        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        public static TDist UseAverageLogarithm<TDist>([IsReturned] TDist to_marginal)
        {
            return to_marginal;
        }

        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        public static TDist DefAverageLogarithm<TDist>([IsReturned] TDist to_marginal)
        {
            return to_marginal;
        }
    }

    /// <summary>Provides outgoing messages for <see cref="Factor.VariablePoint{T}(T, out T)" />, given random arguments to the function.</summary>
    /// <typeparam name="T">The type of the variable.</typeparam>
    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = false)]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp<T> : VariablePointOpBase
    {
        /// <summary>EP message to <c>marginal</c>.</summary>
        /// <param name="use">Incoming message from <c>use</c>.</param>
        /// <param name="def">Incoming message from <c>def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>marginal</c> as the random arguments are varied. The formula is <c>proj[p(marginal) sum_(use,def) p(use,def) factor(use,def,marginal)]/p(marginal)</c>.</para>
        /// </remarks>
        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        [SkipIfAllUniform]
        public static TDist MarginalAverageConditional<TDist>([NoInit] TDist use, TDist def, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMode<T>
        {
            result.SetToProduct(def, use);
            result.Point = result.GetMode();
            return result;
        }

        /// <summary>VMP message to <c>marginal</c>.</summary>
        /// <param name="use">Incoming message from <c>use</c>.</param>
        /// <param name="def">Incoming message from <c>def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the exponential of the average log-factor value, where the average is over all arguments except <c>marginal</c>. The formula is <c>exp(sum_(use,def) p(use,def) log(factor(use,def,marginal)))</c>.</para>
        /// </remarks>
        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        [SkipIfAllUniform]
        public static TDist MarginalAverageLogarithm<TDist>([NoInit] TDist use, TDist def, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMode<T>
        {
            return MarginalAverageConditional(use, def, result);
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = false)]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_Mean<T> : VariablePointOpBase
    {
        /// <summary>EP message to <c>marginal</c>.</summary>
        /// <param name="use">Incoming message from <c>use</c>.</param>
        /// <param name="def">Incoming message from <c>def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is a distribution matching the moments of <c>marginal</c> as the random arguments are varied. The formula is <c>proj[p(marginal) sum_(use,def) p(use,def) factor(use,def,marginal)]/p(marginal)</c>.</para>
        /// </remarks>
        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        [SkipIfAllUniform]
        public static TDist MarginalAverageConditional<TDist>([NoInit] TDist use, TDist def, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMean<T>
        {
            result.SetToProduct(def, use);
            result.Point = result.GetMean();
            return result;
        }

        /// <summary>VMP message to <c>marginal</c>.</summary>
        /// <param name="use">Incoming message from <c>use</c>.</param>
        /// <param name="def">Incoming message from <c>def</c>.</param>
        /// <param name="result">Modified to contain the outgoing message.</param>
        /// <returns>
        ///   <paramref name="result" />
        /// </returns>
        /// <remarks>
        ///   <para>The outgoing message is the exponential of the average log-factor value, where the average is over all arguments except <c>marginal</c>. The formula is <c>exp(sum_(use,def) p(use,def) log(factor(use,def,marginal)))</c>.</para>
        /// </remarks>
        /// <typeparam name="TDist">The type of the marginal of the variable.</typeparam>
        [SkipIfAllUniform]
        public static TDist MarginalAverageLogarithm<TDist>([NoInit] TDist use, TDist def, TDist result)
            where TDist : SettableToProduct<TDist>, HasPoint<T>, CanGetMean<T>
        {
            return MarginalAverageConditional(use, def, result);
        }
    }

    /// <summary>
    /// Implements the Rprop root-finding algorithm.
    /// </summary>
    /// <remarks>
    /// Reference: 
    /// "A Direct Adaptive Method for Faster Backpropagation Learning: The RPROP Algorithm"
    /// Martin Riedmiller, Heinrich Braun
    /// 1993
    /// </remarks>
    public class RpropBufferData
    {
        public static bool debug;
        bool hasPrevious;
        double prevPoint;
        public double nextPoint;
        double scaleUp = 2;
        double stepsize;
        public double lowerBound = double.NegativeInfinity, upperBound = double.PositiveInfinity;

        public void SetNextPoint(double currPoint, double currDeriv)
        {
            if (double.IsNaN(currPoint)) throw new ArgumentException("currPoint is NaN");
            if (double.IsInfinity(currPoint)) throw new ArgumentException("currPoint is infinite");
            if (double.IsNaN(currDeriv)) throw new ArgumentException("currDeriv is NaN");
            if (double.IsInfinity(currDeriv)) throw new ArgumentException("currDeriv is infinite");
            if (hasPrevious)
            {
                double prevStep = currPoint - prevPoint;
                double prevSign = Math.Sign(prevStep);
                double currSign = Math.Sign(currDeriv);
                if (prevSign != 0 && currSign != 0)
                {
                    if (prevSign != currSign)
                    {
                        scaleUp = 1.2;
                        stepsize /= 2;
                    }
                    else
                    {
                        // prevSign == currSign
                        stepsize *= scaleUp;
                    }
                }
            }
            else
            {
                if (currPoint != 0)
                    stepsize = Math.Abs(currPoint) * 1e-4;
                //else if (currDeriv != 0)
                //    stepsize = Math.Abs(currDeriv);  // this is too unstable in practice
                else
                    stepsize = 1e-4;
            }
            prevPoint = currPoint;
            hasPrevious = true;
            currPoint = Math.Max(currPoint, lowerBound);
            currPoint = Math.Min(currPoint, upperBound);
            while (stepsize > 0)
            {
                nextPoint = currPoint + Math.Sign(currDeriv) * stepsize;
                if(debug)
                    Debug.WriteLine("nextPoint = {0}, currPoint = {1}, currDeriv = {2}, stepsize = {3}", nextPoint, currPoint, currDeriv, stepsize);
                if (nextPoint >= lowerBound && nextPoint <= upperBound)
                    break;
                stepsize /= 2;
            }
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = true)]
    [Buffers("buffer")]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_Rprop : VariablePointOpBase
    {
        [Skip]
        public static RpropBufferData BufferInit()
        {
            return new RpropBufferData();
        }

        // must compute new marginal here since Marginal method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static RpropBufferData Buffer([NoInit] Gaussian use, Gaussian def, Gaussian to_marginal, RpropBufferData buffer)
        {
            var currDist = use * def;
            if (currDist.IsPointMass)
            {
                buffer.nextPoint = currDist.Point;
                return buffer;
            }
            // cannot use buffer.nextPoint as currPoint since Marginal could be initialized by user
            double currPoint;
            if (to_marginal.IsPointMass)
            {
                currPoint = to_marginal.Point;
            }
            else
            {
                currPoint = currDist.GetMean();
            }
            // deriv of -0.5*prec*x^2+pm*x
            // is -prec*x + pm
            double currDeriv = currDist.MeanTimesPrecision - currDist.Precision * currPoint;
            buffer.SetNextPoint(currPoint, currDeriv);
            return buffer;
        }

        public static Gaussian MarginalAverageConditional([IgnoreDependency] Gaussian use, [IgnoreDependency] Gaussian def, [RequiredArgument] RpropBufferData buffer, Gaussian result)
        {
            result.Point = buffer.nextPoint;
            return result;
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = true)]
    [Buffers("bufferTG")]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_RpropTruncatedGaussian : VariablePointOpBase
    {
        [Skip]
        public static RpropBufferData BufferTGInit()
        {
            return new RpropBufferData();
        }

        // must compute new marginal here since Marginal method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static RpropBufferData BufferTG([NoInit] TruncatedGaussian use, TruncatedGaussian def, TruncatedGaussian to_marginal, RpropBufferData bufferTG)
        {
            var currDist = use * def;
            if (currDist.IsPointMass)
            {
                bufferTG.nextPoint = currDist.Point;
                return bufferTG;
            }
            // cannot use buffer.nextPoint as currPoint since Marginal could be initialized by user
            double currPoint;
            if (to_marginal.IsPointMass)
            {
                currPoint = to_marginal.Point;
            }
            else
            {
                currPoint = currDist.GetMean();
            }
            // deriv of -0.5*prec*x^2+pm*x
            // is -prec*x + pm
            double currDeriv = currDist.Gaussian.MeanTimesPrecision - currDist.Gaussian.Precision * currPoint;
            bufferTG.lowerBound = currDist.LowerBound;
            bufferTG.upperBound = currDist.UpperBound;
            bufferTG.SetNextPoint(currPoint, currDeriv);
            return bufferTG;
        }

        public static TruncatedGaussian MarginalAverageConditional([IgnoreDependency] TruncatedGaussian use, [IgnoreDependency] TruncatedGaussian def, [RequiredArgument] RpropBufferData bufferTG, TruncatedGaussian result)
        {
            result.Point = bufferTG.nextPoint;
            return result;
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = true)]
    [Buffers("bufferTGa")]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_RpropTruncatedGamma : VariablePointOpBase
    {
        [Skip]
        public static RpropBufferData BufferTGaInit()
        {
            var result = new RpropBufferData();
            if (!VariablePointOp_RpropGamma.UseMean)
                result.lowerBound = 0;
            return result;
        }

        // must compute new marginal here since Marginal method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static RpropBufferData BufferTGa([NoInit] TruncatedGamma use, TruncatedGamma def, TruncatedGamma to_marginal, RpropBufferData bufferTGa)
        {
            var currDist = use * def;
            if (currDist.IsPointMass)
            {
                if (VariablePointOp_RpropGamma.UseMean)
                    bufferTGa.nextPoint = Math.Log(currDist.Point);
                else
                    bufferTGa.nextPoint = currDist.Point;
                return bufferTGa;
            }
            // cannot use buffer.nextPoint as currPoint since Marginal could be initialized by user
            double currPoint;
            if (to_marginal.IsPointMass)
            {
                currPoint = to_marginal.Point;
            }
            else
            {
                currPoint = currDist.GetMean();
            }
            double currDeriv, currDeriv2;
            if (VariablePointOp_RpropGamma.UseMean)
            {
                bufferTGa.lowerBound = Math.Log(currDist.LowerBound);
                bufferTGa.upperBound = Math.Log(currDist.UpperBound);
                currDeriv = currDist.Gamma.Shape - currDist.Gamma.Rate * currPoint;
                if (currPoint <= 0)
                    throw new ArgumentException($"currPoint ({currPoint}) <= 0");
                bufferTGa.SetNextPoint(Math.Log(currPoint), currDeriv);
            }
            else
            {
                bufferTGa.lowerBound = currDist.LowerBound;
                bufferTGa.upperBound = currDist.UpperBound;
                currDist.Gamma.GetDerivatives(currPoint, out currDeriv, out currDeriv2);
                bufferTGa.SetNextPoint(currPoint, currDeriv);
            }
            return bufferTGa;
        }

        public static TruncatedGamma MarginalAverageConditional([IgnoreDependency] TruncatedGamma use, [IgnoreDependency] TruncatedGamma def, [RequiredArgument] RpropBufferData bufferTGa, TruncatedGamma result)
        {
            if (VariablePointOp_RpropGamma.UseMean)
            {
                result.Point = Math.Exp(bufferTGa.nextPoint);
            }
            else
            {
                result.Point = bufferTGa.nextPoint;
            }
            return result;
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = true)]
    [Buffers("buffer0")]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_RpropGamma : VariablePointOpBase
    {
        public static bool UseMean;

        [Skip]
        public static RpropBufferData Buffer0Init()
        {
            var result = new RpropBufferData();
            if (!UseMean)
                result.lowerBound = 0;
            return result;
        }

        // must compute new marginal here since Marginal method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static RpropBufferData Buffer0([NoInit] Gamma use, Gamma def, Gamma to_marginal, RpropBufferData buffer0)
        {
            var currDist = use * def;
            if (currDist.IsPointMass)
            {
                if(UseMean)
                    buffer0.nextPoint = Math.Log(currDist.Point);
                else
                    buffer0.nextPoint = currDist.Point;
                return buffer0;
            }
            // cannot use buffer.nextPoint as currPoint since Marginal could be initialized by user
            double currPoint;
            if (to_marginal.IsPointMass)
            {
                currPoint = to_marginal.Point;
            }
            else
            {
                currPoint = currDist.GetMean();
            }
            double currDeriv, currDeriv2;
            if (UseMean)
            {
                currDeriv = currDist.Shape - currDist.Rate * currPoint;
                if (currPoint <= 0)
                    throw new ArgumentException($"currPoint ({currPoint}) <= 0");
                buffer0.SetNextPoint(Math.Log(currPoint), currDeriv);
            }
            else
            {
                currDist.GetDerivatives(currPoint, out currDeriv, out currDeriv2);
                buffer0.SetNextPoint(currPoint, currDeriv);
            }
            return buffer0;
        }

        public static Gamma MarginalAverageConditional([IgnoreDependency] Gamma use, [IgnoreDependency] Gamma def, [RequiredArgument] RpropBufferData buffer0, Gamma result)
        {
            if (UseMean)
            {
                result.Point = Math.Exp(buffer0.nextPoint);
            }
            else
            {
                result.Point = buffer0.nextPoint;
            }
            return result;
        }
    }

    [FactorMethod(typeof(Factor), "VariablePoint<>", Default = true)]
    [Buffers("bufferBeta")]
    [Quality(QualityBand.Preview)]
    public class VariablePointOp_RpropBeta : VariablePointOpBase
    {
        [Skip]
        public static RpropBufferData BufferBetaInit()
        {
            return new RpropBufferData() { lowerBound = 0, upperBound = 1 };
        }

        // must compute new marginal here since Marginal method cannot modify buffer
        // Buffer must be called once per iter, so cannot be fresh or have triggers
        [SkipIfAllUniform]
        public static RpropBufferData BufferBeta([NoInit] Beta use, Beta def, Beta to_marginal, RpropBufferData bufferBeta)
        {
            var currDist = use * def;
            if (currDist.IsPointMass)
            {
                bufferBeta.nextPoint = currDist.Point;
                return bufferBeta;
            }
            // cannot use buffer.nextPoint as currPoint since Marginal could be initialized by user
            double currPoint;
            if (to_marginal.IsPointMass)
            {
                currPoint = to_marginal.Point;
            }
            else
            {
                currPoint = currDist.GetMean();
            }
            double currDeriv, currDeriv2;
            currDist.GetDerivatives(currPoint, out currDeriv, out currDeriv2);
            bufferBeta.SetNextPoint(currPoint, currDeriv);
            return bufferBeta;
        }

        public static Beta MarginalAverageConditional([IgnoreDependency] Beta use, [IgnoreDependency] Beta def, [RequiredArgument] RpropBufferData bufferBeta, Beta result)
        {
            result.Point = bufferBeta.nextPoint;
            return result;
        }
    }
}
